"""
===========================
Hill5 analytic infall model 
===========================
.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>

Code translated from:
https://bitbucket.org/devries/analytic_infall/overview

Original source:
http://adsabs.harvard.edu/abs/2005ApJ...620..800D
"""
import numpy as np
import model

def hill5_model( xarr, tau, v_lsr,  v_infall,  sigma,  tpeak, TBG=2.73):
    """
    The rest of this needs to be translated from C
    """
  
    vf = v_lsr+v_infall
    vr = v_lsr-v_infall
  
    velocity_array = np.array(xarr.as_unit('km/s'))
    tauf = tau*np.exp(-((velocity_array-vf)/sigma)**2 / 2.0 )
    taur = tau*np.exp(-((velocity_array-vr)/sigma)**2 / 2.0 )

    subf = ( (1-np.exp(-tauf))/tauf * (tauf>1e-4) + (tauf < 1e-4) )
    subr = ( (1-np.exp(-taur))/taur * (taur>1e-4) + (taur < 1e-4) )
    subf[subf != subf] = 1.0
    subr[subr != subr] = 1.0
  
    frequency = np.array(xarr.as_unit('Hz'))
    hill_array =(jfunc(tpeak,frequency)-jfunc(TBG,frequency))*(subf-np.exp(-tauf)*subr)

    if np.isnan(hill_array).any():
        raise ValueError("Hill5 model has a NAN")

    return hill_array

def jfunc(t, nu):
    """
    t- kelvin
    nu - Hz?
    """


    H=6.6260755e-27
    K=1.380658e-16
    NSIG=2.0

    # I think this is only necessary for C
    #if(nu<1.0e-6) return t
    
    to = H*nu/K
    return to/(np.exp(to/t)-1.0)

hill5_fitter = model.SpectralModel(hill5_model, 5,
        parnames=['tau', 'v_lsr',  'v_infall',  'sigma', 'tpeak'], 
        parlimited=[(True,False),(False,False),(True,False),(True,False), (True,False)], 
        parlimits=[(0,0), (0,0), (0,0), (0,0), (0,0)],
        shortvarnames=("\\tau","v_{lsr}","v_{infall}","\\sigma","T_{peak}"), # specify the parameter names (TeX is OK)
        fitunits='Hz' )



"""
"""
from classes import Spectrum,Spectra,ObsBlock
import fitters,plotters,baseline,units
import smooth
import correlate
import headers
import logger
from .. import config
import moments
import units

import readers
def register_reader(filetype, function, suffix, default=False):
    ''' 
    Register a reader function.

    Required Arguments:

        *filetype*: [ string ]
            The file type name

        *function*: [ function ]
            The reader function.  Should take a filename as input
            and return an X-axis object (see units.py), a spectrum,
            an error spectrum (initialize it to 0's if empty), and a
            pyfits header instance

        *suffix*: [ int ]
            What suffix should the file have?

    Optional Keyword Arguments:

    '''

    readers.readers[filetype] = function
    if suffix in readers.suffix_types:
        if default:
            readers.suffix_types[suffix].insert(0,filetype)
        else:
            readers.suffix_types[suffix].append(filetype)
    else: # if it's the first, it defaults to default!
        readers.suffix_types[suffix] = [filetype]

register_reader('fits',readers.open_1d_fits,'fits',default=True)
register_reader('fits',readers.open_1d_fits,'fit')
register_reader('pyfits',readers.open_1d_pyfits,'')
register_reader('txt',readers.open_1d_txt,'txt')
register_reader('txt',readers.open_1d_txt,'dat')
register_reader('tspec',readers.tspec_reader,'fits')
register_reader('hdf5',readers.open_hdf5,'hdf5')

import writers
def register_writer(filetype, function, suffix, default=False):
    ''' 
    Register a writer function.

    Required Arguments:

        *filetype*: [ string ]
            The file type name

        *function*: [ function ]
            The writer function.  Will be an attribute of Spectrum
            object, and called as spectrum.Spectrum.write_hdf5(), 
            for example. 

        *suffix*: [ int ]
            What suffix should the file have?

    Optional Keyword Arguments:

    '''

    writers.writers[filetype] = function
    if suffix in writers.suffix_types:
        if default:
            writers.suffix_types[suffix].insert(0,filetype)
        else:
            writers.suffix_types[suffix].append(filetype)
    else: # if it's the first, it defaults to default!
        writers.suffix_types[suffix] = [filetype]

register_writer('hdf5',writers.write_hdf5,'hdf5',default=False)
register_writer('fits',writers.write_fits,'fits',default=True)
#register_writer('txt',writers.write_ascii,'txt')
#register_writer('txt',writers.write_ascii,'txt')



try:
    import pyslalib

    def topo_to_geo(ra,dec,latitude,lst,height=None):
        """
        :ra: Decimal right ascension (degrees)
        :dec: Decimal declination (degrees)
        :latitude: Observatory latitude in degrees
        :lst: Local sidereal time (hours)

        .. warning: this option was implemented but returns inconsistent results
            and has therefore been disabled
        :height: (optional) Height above the reference spheroid in meters
            (if height is specified, a more precise version of the code is used)
        """
        import pyslalib
        ra_rad = ra*np.pi/180.
        dec_rad = dec*np.pi/180.
        latitude_rad = latitude*np.pi/180.
        lst_rad = lst*15.*np.pi/180.
        if height is None:
            return pyslalib.slalib.sla_rverot(ra_rad,dec_rad,latitude_rad,lst_rad)
        else:
            return pyslalib.slalib.sla_rverot(ra_rad,dec_rad,latitude_rad,lst_rad)
            vearth_6d = pyslalib.slalib.sla_pvobs(latitude_rad, height, lst_rad)
            vradec_3d = pyslalib.slalib.sla_dcs2c(ra_rad, dec_rad)
            projected_vearth = -1*pyslalib.slalib.sla_dvdv(vearth_6d[3:],vradec_3d)
            return projected_vearth * astronomical_unit_cm*length_dict['cm']/length_dict['km']


    def geo_to_bary(ra,dec,jd,epoch):
        """
        For a given ra/dec, return the conversion from geocentric velocity to heliocentric

        :ra: Decimal right ascension (degrees)
        :dec: Decimal declination (degrees)
        :jd: Modified (2000 = 51544) julian date
        :epoch: Epoch of observations (e.g., 2000, 2008.202)

        follows instructions given here:
        http://star-www.rl.ac.uk/docs/sun67.htx/node230.html

        *  Star vector, J2000
            CALL sla_DCS2C(RM,DM,V)

        *  Earth/Sun velocity and position, J2000
            CALL sla_EVP(TDB,2000D0,DVB,DPB,DVH,DPH)

        *  Radial velocity correction due to Earth orbit (km/s)
            VCORB = -sla_DVDV(V,DVH)*149.597870D6

        """
        import pyslalib
        velocity_3d = pyslalib.slalib.sla_dcs2c(ra*np.pi/180.,dec*np.pi/180.)
        dvb, dpb, dvh, dph = pyslalib.slalib.sla_evp(jd,epoch) 

        vcorb = -pyslalib.slalib.sla_dvdv(velocity_3d,dvb)*149.597870e6

        return vcorb

    def geo_to_helio(ra,dec,jd,epoch):
        """
        For a given ra/dec, return the conversion from geocentric velocity to heliocentric

        :ra: Decimal right ascension (degrees)
        :dec: Decimal declination (degrees)
        :jd: Modified (2000 = 51544) julian date
        :epoch: Epoch of observations (e.g., 2000, 2008.202)

        follows instructions given here:
        http://star-www.rl.ac.uk/docs/sun67.htx/node230.html

        *  Star vector, J2000
            CALL sla_DCS2C(RM,DM,V)

        *  Earth/Sun velocity and position, J2000
            CALL sla_EVP(TDB,2000D0,DVB,DPB,DVH,DPH)

        *  Radial velocity correction due to Earth orbit (km/s)
            VCORB = -sla_DVDV(V,DVH)*149.597870D6

        """
        import pyslalib
        velocity_3d = pyslalib.slalib.sla_dcs2c(ra*np.pi/180.,dec*np.pi/180.)
        dvb, dpb, dvh, dph = pyslalib.slalib.sla_evp(jd,epoch) 

        vcorb = -pyslalib.slalib.sla_dvdv(velocity_3d,dvh)*149.597870e6

        return vcorb

    def helio_to_lsr(ra, dec):
        """
        """

        import pyslalib
        return pyslalib.slalib.sla_rvlsrk(ra/180.*np.pi, dec/180.*np.pi)

    def topo_to_lsr(ra, dec, latitude, lst, jd, epoch, height=None):
        print "helio->lsr: ",helio_to_lsr(ra,dec)
        print "geo->helio: ",geo_to_helio(ra,dec,jd,epoch)
        print "topo->geo: ",topo_to_geo(ra,dec,latitude,lst,height=height)
        return helio_to_lsr(ra,dec) + geo_to_helio(ra,dec,jd,epoch) + topo_to_geo(ra,dec,latitude,lst,height=height)

    def frame_grid(ra=0.0,dec=0.0,jd=51544,lst=0,vtopo=0.0,epoch=2000,):
        frame_names = ['TOPO','GEO','BARY','HELI','LSRK']
        frame1_to_frame2 = dict([(n1,None) for n1 in frame_names])
        for frame1 in frame_names:
            frame1_to_frame2[frame1] = dict([(n2,None) for n2 in frame_names])
            for frame2 in frame_names:
                if frame1 == frame2:
                    frame1_to_frame2[frame1][frame2] = 0.0
                elif frame1 == 'TOPO' and frame2 == 'GEO':
                    frame1_to_frame2[frame1][frame2] = topo_to_geo(ra,dec,latitude,lst)


except ImportError:
    pass

if __name__ == "__main__":
    import pytest 
    @pytest.mark.parametrize
    def test_vlsr(ra, dec, utyr, utmon, utday, epoch, latitude, elevation):
        pass




"""
==================
NH3 fitter wrapper
==================

Wrapper to fit ammonia spectra.  Generates a reasonable guess at the position and velocity using a gaussian fit
"""
import pyspeckit
from matplotlib import pyplot 
import copy
import random

title_dict = {'oneone':'NH$_3(1,1)$', 
'twotwo':'NH$_3(2,2)$', 
'threethree':'NH$_3(3,3)$', 
'fourfour':'NH$_3(4,4)$', }

def fitnh3tkin(input_dict, dobaseline=True, baselinekwargs={}, crop=False, guessline='twotwo',
        tex=15,tkin=20,column=15.0,fortho=0.66, tau=None, thin=False, quiet=False, doplot=True, fignum=1,
        guessfignum=2, smooth=False, scale_keyword=None, rebase=False, npeaks=1, guesses=None,
        **kwargs): 
    """
    Given a dictionary of filenames and lines, fit them together
    e.g. {'oneone':'G000.000+00.000_nh3_11.fits'}
    """
    spdict = dict([ (linename,pyspeckit.Spectrum(value, scale_keyword=scale_keyword)) 
        if type(value) is str else (linename,value) for linename, value in input_dict.iteritems() ])
    splist = spdict.values()

    for sp in splist: # required for plotting, cropping
        sp.xarr.convert_to_unit('km/s')

    if crop and len(crop) == 2:
        for sp in splist:
            sp.crop(*crop)

    if dobaseline:
        for sp in splist:
            sp.baseline(**baselinekwargs)

    if smooth and type(smooth) is int:
        for sp in splist:
            sp.smooth(smooth)

    spdict[guessline].specfit(fittype='gaussian', negamp=False, vheight=False)
    ampguess,vguess,widthguess = spdict[guessline].specfit.modelpars
    if widthguess < 0:
        raise ValueError("Width guess was < 0.  This is impossible.")
    print "RMS guess (errspec): ",spdict[guessline].specfit.errspec.mean()
    print "RMS guess (residuals): ",spdict[guessline].specfit.residuals.std()
    errguess = spdict[guessline].specfit.residuals.std()

    if rebase:
        # redo baseline subtraction excluding the centroid +/- about 20 km/s 
        vlow = spdict[guessline].specfit.modelpars[1]-(19.8+spdict[guessline].specfit.modelpars[2]*2.35)
        vhigh = spdict[guessline].specfit.modelpars[1]+(19.8+spdict[guessline].specfit.modelpars[2]*2.35)
        for sp in splist:
            sp.baseline(exclude=[vlow,vhigh], **baselinekwargs)
    
    for sp in splist:
        sp.error[:] = errguess

    spdict[guessline].plotter(figure=guessfignum)
    spdict[guessline].specfit.plot_fit()

    spectra = pyspeckit.Spectra(splist)
    spectra.specfit.npeaks = npeaks

    if tau is not None:
        if guesses is None:
            guesses = [a for i in xrange(npeaks) for a in (tkin+random.random()*i,tex,tau+random.random()*i,widthguess+random.random()*i,vguess+random.random()*i,fortho)]
        spectra.specfit(fittype='ammonia_tau',quiet=quiet,multifit=True,guesses=guesses, thin=thin, **kwargs)
    else:
        if guesses is None:
            guesses = [a for i in xrange(npeaks) for a in (tkin+random.random()*i,tex,column+random.random()*i,widthguess+random.random()*i,vguess+random.random()*i,fortho)]
        spectra.specfit(fittype='ammonia',quiet=quiet,multifit=True,guesses=guesses, thin=thin, **kwargs)

    if doplot:
        plot_nh3(spdict,spectra,fignum=fignum)

    return spdict,spectra

def plot_nh3(spdict,spectra,fignum=1, show_components=False, residfignum=None, **plotkwargs):
    """
    Plot the results from a multi-nh3 fit
    """ 
    spectra.plotter.figure = pyplot.figure(fignum)
    spectra.plotter.axis = spectra.plotter.figure.gca()
    pyplot.clf()
    splist = spdict.values()

    for sp in splist:
        sp.xarr.convert_to_unit('km/s',quiet=True)
        sp.specfit.fitter = copy.copy(spectra.specfit.fitter)
        sp.specfit.modelpars = spectra.specfit.modelpars
        sp.specfit.npeaks = spectra.specfit.npeaks
        sp.specfit.fitter.npeaks = spectra.specfit.npeaks
        if spectra.specfit.modelpars is not None:
            sp.specfit.model = sp.specfit.fitter.n_ammonia(pars=spectra.specfit.modelpars, parnames=spectra.specfit.fitter.parnames)(sp.xarr)

    if len(splist) == 2:
        axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(212) }
    elif len(splist) == 3:
        axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(223), 'threethree':pyplot.subplot(224), 'fourfour':pyplot.subplot(224) }
    elif len(splist) == 4:
        axdict = { 'oneone':pyplot.subplot(221), 'twotwo':pyplot.subplot(222), 'threethree':pyplot.subplot(223), 'fourfour':pyplot.subplot(224) }
    for linename,sp in spdict.iteritems():
        sp.plotter.axis=axdict[linename] # permanent
        sp.plotter(axis=axdict[linename],title=title_dict[linename], **plotkwargs)
        sp.specfit.Spectrum.plotter = sp.plotter
        sp.specfit.selectregion(reset=True)
        if sp.specfit.modelpars is not None:
            sp.specfit.plot_fit(annotate=False, show_components=show_components)
    if spdict['oneone'].specfit.modelpars is not None:
        spdict['oneone'].specfit.annotate(labelspacing=0.05,prop={'size':'small','stretch':'extra-condensed'},frameon=False)

    if residfignum is not None:
        pyplot.figure(residfignum)
        pyplot.clf()
        if len(splist) == 2:
            axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(212) }
        elif len(splist) == 3:
            axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(223), 'threethree':pyplot.subplot(224), 'fourfour':pyplot.subplot(224) }
        elif len(splist) == 4:
            axdict = { 'oneone':pyplot.subplot(221), 'twotwo':pyplot.subplot(222), 'threethree':pyplot.subplot(223), 'fourfour':pyplot.subplot(224) }
        for linename,sp in spdict.iteritems():
            sp.specfit.plotresiduals(axis=axdict[linename])



def fitnh3(spectrum, vrange=[-100,100], vrangeunits='km/s', quiet=False,
        Tex=20,Tkin=15,column=1e15,fortho=1.0, tau=None): 

    if vrange:
        spectrum.xarr.convert_to_unit(vrangeunits)
        spectrum.crop(*vrange, units=vrangeunits)

    spectrum.specfit(fittype='gaussian',negamp=False)
    ampguess,vguess,widthguess = spectrum.specfit.modelpars

    if tau is None:
        spectrum.specfit(fittype='ammonia',quiet=quiet,multifit=True,guesses=[Tex,Tkin,column,widthguess,vguess,fortho])
    else:
        spectrum.specfit(fittype='ammonia_tau',quiet=quiet,multifit=True,guesses=[Tex,Tkin,tau,widthguess,vguess,fortho])

    return spectrum


def BigSpectrum_to_NH3dict(sp, vrange=None):
    """
    A rather complicated way to make the spdicts above given a spectrum...
    """

    spdict = {}
    for linename,freq in pyspeckit.spectrum.models.ammonia.freq_dict.iteritems():
        if vrange is not None:
            freq_test_low  = freq - freq * vrange[0]/pyspeckit.units.speedoflight_kms
            freq_test_high = freq - freq * vrange[1]/pyspeckit.units.speedoflight_kms
        else:
            freq_test_low = freq_test_high = freq

        if (sp.xarr.as_unit('Hz').in_range(freq_test_low) or
                sp.xarr.as_unit('Hz').in_range(freq_test_high)):
            spdict[linename] = sp.copy()
            spdict[linename].xarr.convert_to_unit('GHz')
            spdict[linename].xarr.refX = freq
            spdict[linename].xarr.refX_units = 'Hz'
            #spdict[linename].baseline = copy.copy(sp.baseline)
            #spdict[linename].baseline.Spectrum = spdict[linename]
            #spdict[linename].specfit = copy.copy(sp.specfit)
            #spdict[linename].specfit.fitter = copy.copy(sp.specfit.fitter)
            #spdict[linename].specfit.Spectrum = spdict[linename]
            #spdict[linename].specfit._full_model()
            spdict[linename].xarr.convert_to_unit('km/s')
            if vrange is not None:
                try:
                    spdict[linename].crop(*vrange, units='km/s')
                except IndexError:
                    # if the freq in range, but there's no data in range, remove
                    spdict.pop(linename)

    return spdict
                    
def plotter_override(sp, vrange=None, **kwargs):
    """
    Do plot_nh3 with syntax similar to plotter()
    """

    spdict = BigSpectrum_to_NH3dict(sp, vrange=vrange)

    if len(spdict) not in (2,3,4):
        raise ValueError("Not enough lines; don't need to use the NH3 plot wrapper")

    plot_nh3(spdict, sp, **kwargs)



from pyspeckit.specwarnings import warn
try:
    import coords
    coordsOK = True
except ImportError:
    coordsOK = False
from numpy import pi,arctan2,sin,cos,tan

def posang(l1,b1,l2,b2,system='galactic',**kwargs):
    """
    Return the position angle between two points assuming a rectilinear
    coordinate system (I think; at the very least I am making no corrections
    for wcs).

    INPUT:
    longitude1, latitude1, longitude2, latitude2

    Defaults to GALACTIC coordinates.  **kwargs are passed to coords.Position
    """
    if not coordsOK:
        raise ImportError( "posang requires coords." )
    pos1 = coords.Position([l1,b1],system=system,**kwargs)
    ra1,dec1 = pos1.j2000()
    pos2 = coords.Position([l2,b2],system=system,**kwargs)
    ra2,dec2 = pos2.j2000()

    radiff  = (ra1-ra2)/180.*pi

    angle = arctan2( sin(radiff) , cos(dec1*pi/180.)*tan(dec2*pi/180.) - sin(dec1*pi/180.)*cos(radiff) ) /pi*180

    return angle



"""
==========
Model Grid
==========
Fit a line based on parameters output from a grid of models
"""
import numpy as np
from pyspeckit.mpfit import mpfit
import matplotlib.cbook as mpcb
import copy
try:
    import scipy.interpolate
    import scipy.ndimage
    scipyOK = True
except ImportError:
    scipyOK=False

def gaussian_line(xax, maxamp, tau, offset, width):
    """
    A Gaussian line function in which the 
    """
    return np.exp(-(xax-offset)**2/(2.0*width**2)) * maxamp * (1.0-np.exp(-1*tau))

def line_params_2D(gridval1, gridval2, griddim1, griddim2, valuegrid):
    """
    Given a 2D grid of modeled line values - the amplitude, e.g. excitation temperature,
    and the optical depth, tau - return the model spectrum

    griddims contains the names of the axes and their values... it should have the same 
    number of entries as gridpars
    """

    if not scipyOK:
        raise ImportError("Scipy could not be imported, therefore interpolation is not available.")

    #matchpt1 = np.argmin( np.abs( gridval1 - griddim1[0,:] ))
    #matchpt2 = np.argmin( np.abs( gridval2 - griddim2[:,0] ))

    return scipy.ndimage.map_coordinates(valuegrid,np.array([[gridval2],[gridval1]]),order=1)

    interpgrid = scipy.interpolate.interp2d( 
            griddim1[ gridval1-5:gridval1+5, gridval2-5:gridval2+5].ravel(),
            griddim2[ gridval1-5:gridval1+5, gridval2-5:gridval2+5].ravel(),
            valuegrid[gridval1-5:gridval1+5, gridval2-5:gridval2+5].ravel())
    
    return interpgrid(gridval1,gridval2)

def line_model_2par(xax, center, width, gridval1, gridval2, griddim1, griddim2, maxampgrid, taugrid,
        linefunction=gaussian_line):
    """
    Returns the spectral line that matches the given x-axis

    xax, center, width must be in the same units!
    """

    maxamp = line_params_2D(gridval1, gridval2, griddim1, griddim2, maxampgrid)
    tau = line_params_2D(gridval1, gridval2, griddim1, griddim2, taugrid)

    return linefunction(xax,maxamp,tau,offset,width)





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# This module implements the Spectrum1D class. It is eventually supposed to migrate to astropy core

# Python packages
from abc import ABCMeta, abstractmethod, abstractproperty
from copy import copy, deepcopy

# External packages
import numpy as np  

# Local packages
from astropy.nddata import NDData

class SpectrumErrorArray(object):
    """ Represents the error axis of a Spectrum object
    
    .. warning::
        This is skeleton code!
    
    Spectral data errors can be represented in a variety of ways, e.g. 1-sigma
    errors, inverse variance, variance, etc. In order to provide generic functionality,
    the error data must carry with it some metadata and operations that help
    handle transformations of the error array.

    Parameters
    ----------
    array : `~numpy.ndarray`
        The error values
    
    Notes
    -----
    This class is abstract and can only be subclassed and used by
    classes that will access its API.

    """
    
    # This is an abstract class and can only be subclassed.
    __metaclass__ = ABCMeta
    
    def __init__(self, values):
        self.values = values
    
    @abstractmethod
    def interpolate(self, new_dispersion):
        pass

class InverseVarianceErrorArray(SpectrumErrorArray):
    # ...
    
    def interpolate(self, new_dispersion):
        pass

class VarianceErrorArray(SpectrumErrorArray):
    # ...
    
    def interpolate(self, new_dispersion):
        pass

class Spectrum1DBase(NDData):
    """ Base class for Spectrum1D objects.
    
    .. warning::
        This is skeleton code!
    
    `Spectrum1DBase` is a stripped-down superclass of the Spectrum1D object
    that allows for creation of a Spectrum-like object **without** an internal
    dispersion array. This allows for the possibility of creating a 
    SpectrumCollection type class that could contain many spectra that share
    the same dispersion axis (e.g. SDSS spectra from the same plate).

    Parameters
    ----------
    flux : `~numpy.ndarray`
        The flux data as an array
        
    Notes
    -----
    This class should be 'private' in that it should really only be subclassed
    and used by Spectrum1D and SpectrumCollection

    """
    def __init__(self):
        pass # not yet implemented
        
    @property
    def spectrum1DBase(self):
        """ This method returns an object equivalent to this spectrum but as
            a Spectrum1DBase object, i.e. without a dispersion array.
            It is left to subclasses to handle this in more detail (e.g., see Spectrum1D).
        """
        return self    
    
class Spectrum1D(Spectrum1DBase):
    """ Class for 1-dimensional Spectrum objects.
    
    .. warning::
        This is skeleton code!
    
    `Spectrum1D` provides a container for 1-dimensional spectral data as well
    as generic operations for this data.

    Parameters
    ----------
    flux : `~numpy.ndarray`
        The flux data as an array
    dispersion : `~numpy.ndarray`
        An array of data representing the dispersion axis of the spectral data,
        e.g. wavelength, frequency, energy, velocity, etc.
    <NDData parameters...>
        
    Notes
    -----
    
    """
    # This is an abstract class and can only be subclassed.
    __metaclass__ = ABCMeta
    
    @property
    def spectrum1DBase(self):
        """ Return a new Spectrum1DBase object from this spectrum.
        
            This basically is the same thing but without the dispersion array,
            but we'll see how this develops. """
        # There will be a better way to do this, but this is the general idea.
        # Basically I want to create a Spectrum1DBase object from a Spectrum1D object.
        spectrum_copy = deepcopy(self)
        spectrum_copy.dispersion = None
        return spectrum_copy

    def slice_dispersion(self, start=None, end=None):
        """ Slices the data arrays based on values in the dispersion array
        
            Parameters
            ----------
            start : any numeric type, optional
                The minimum value in the dispersion axis to slice on
            end : any numeric type, optional
                The maximum value in the dispersion axis to slice on
        """
        pass
    
    def slice_pixel(self, start=None, end=None):
        """ Slices the data arrays on pixel index (e.g. array index)
        
            Parameters
            ----------
            start : int, optional
                The starting index to slice the arrays on
            end : int, optional
                The ending index to slice the arrays on
            
            Notes
            -----
            This is equivalent to slicing each array like array[start:end]
            
        """
        pass
    
    def interpolate_linear(self, new_dispersion):
        """ Uses linear interpolation to resample the internal arrays onto
            the specified dispersion axis.
        
            Parameters
            ----------
            new_dispersion : `~numpy.ndarray`
                The new dispersion array to interpolate on to
            
        """
        pass
        
    def interpolate_bspline(self, new_dispersion):
        """ Uses B-spline interpolation to resample the internal arrays onto
            the specified dispersion axis.
        
            Parameters
            ----------
            new_dispersion : `~numpy.ndarray`
                The new dispersion array to interpolate on to
            
        """
        pass
    
    def smooth_boxcar(self, width):
        """ Uses boxcar smoothing to smooth the internal arrays.
        
            Parameters
            ----------
            width : integer
                The boxcar width in pixels.
                
            Reference: <a href="http://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.boxcar.html" target="_blank">scipy.signal.boxcar</a>
        
        """
        # I don't know if this is really the proper way to do this... should be tested!!
        from scipy.signal import convolve, boxcar
        self.flux = convolve(self.flux, boxcar(M=width))
    
    def smooth_custom(self):
        """ Allows for a user defined smoothing operation.
            
            This is simply a template function meant to be overwritten
            by a custom function e.g.:
            
            def so_smooth(self):
                # Do something with flux, dispersion, error
            
            spectrumObject.custom_smooth = so_smooth
            spectrumObject.custom_smooth()
        """
        pass
    
    def plot(self, **kwargs):
        """ Plotting utility for the spectrum.
            
            Parameters
            ----------
            axes : `~matplotlib.pyplot.Axes`, optional
                A matplotlib Axes() object to plot the spectrum on
            filename : str, optional
                A filename to save the plot to
            show_error : bool, optional
                Decide whether to use the error data when plotting
                
            Usage
            -----
            s = Spectrum1D(...)
            
            s.plot() - Attempt to display the object on screen
                       using matplotlib's `show()` command.

            s.plot(axes=ax) - Plot the data onto the specified
                              Axes() object.
            
            s.plot(axes=ax) - Plot the data onto the specified
                              Axes() object.
            
            s.plot(filename="myPlot.png") - Plot the data onto a new figure and
                                save to the specified filename and path
                                (default path if not specified).
                                The format of the file will be deduced from the
                                extension (e.g. ps, png, pdf).
            
            Notes
            -----
            Where it makes sense (i.e. not conflicting), multiple parameters
            can be specified. For example,
            
            s.plot(filename="myPlot.pdf", axes=ax)
            
            will both write a file to the disk and return a matplotlib.Axes() object.
        
        """
        pass
            
class Spectrum1DCollection(object):
    """ A collection object for spectra that share the same dispersion information.
    
    """
    def __init__(self):
        self.spectra = list()
        self.dispersion = None
        
    def append(self, spectrum):
        """ Add a spectrum (of type Spectrum1D or Spectrum1DBase) to this collection.
        
        """
        self.spectra.append(new_spectrum.spectrum1DBase)
    
    def len(self):
        return len(self.spectra)
    
    @property
    def dispersion(self):
        return self.dispersion
    
    @dispersion.setter
    def dispersion(self, new_dispersion):
        """ Set the dispersion array to be used for all spectra in this collection.
        
        The dispersion argument accepts both a numpy array and a Spectrum1D object.
        When the latter is specified, the dispersion is extracted from the object.
        """
        if is_instance(new_dispersion, Spectrum1D):
            self.dispersion = new_dispersion.dispersion
        elif is_instance(new_dispersion, list):
            self.dispersion = np.array(new_dispersion)
        elif is_instance(new_dispersion, np.array):
            self.dispersion = new_dispersion
        else:
            raise ValueError, "The dispersion specified could is not a known type. Should be a list, a numpy array, or a Spectrum1D object."
        # Add other types that unambiguously define a dispersion array.






"""
=================
Lorentzian Fitter
=================

The simplest and most useful model.

Until 12/23/2011, lorentzian fitting used the complicated and somewhat bloated
gaussfitter.py code.  Now, this is a great example of how to make your own
model!

"""
import model
import numpy 

def lorentzian(x,A,dx,w, return_components=False):
    """
    Returns a 1-dimensional lorentzian of form
    A*2*pi*w/((x-dx)**2 + ((w/2)**2))
    
    [amplitude,center,width]

    return_components does nothing but is required by all fitters
    
    """
    x = numpy.array(x) # make sure xarr is no longer a spectroscopic axis
    return A/(2.0*numpy.pi)*w/((x-dx)**2 + (w/2.0)**2)

def lorentzian_fitter(multisingle='multi'):
    """
    Generator for lorentzian fitter class
    """

    myclass =  model.SpectralModel(lorentzian, 3,
            parnames=['amplitude','shift','width'], 
            parlimited=[(False,False),(False,False),(True,False)], 
            parlimits=[(0,0), (0,0), (0,0)],
            shortvarnames=('A',r'\Delta x',r'\sigma'),
            multisingle=multisingle,
            )
    myclass.__name__ = "lorentzian"
    
    return myclass




"""
Fit a line based on parameters output from a grid of RADEX models
"""
import numpy as np
from pyspeckit.mpfit import mpfit
from .. import units
from . import fitter,model
import matplotlib.cbook as mpcb
import copy

class radex_model(object):
    def __init__(self, xarr,  
            grid_vwidth=1.0,
            grid_vwidth_scale=False,
            texgrid=None,
            taugrid=None,
            hdr=None,
            path_to_texgrid='',
            path_to_taugrid='',
            temperature_gridnumber=3,
            debug=False,
            verbose=False,
            modelfunc=None,
            **kwargs):
        """
        Use a grid of RADEX-computed models to make a model line spectrum

        The RADEX models have to be available somewhere.
        OR they can be passed as arrays.  If as arrays, the form should be:
        texgrid = ((minfreq1,maxfreq1,texgrid1),(minfreq2,maxfreq2,texgrid2))

        xarr must be a SpectroscopicAxis instance
        xoff_v, width are both in km/s.  With is 'sigma'

        grid_vwidth is the velocity assumed when computing the grid in km/s
            this is important because tau = modeltau / width (see, e.g., 
            Draine 2011 textbook pgs 219-230)
        grid_vwidth_scale is True or False: False for LVG, True for Sphere


        A modelfunc must be specified.  Model functions should take an xarr and
        a series of keyword arguments corresponding to the line parameters
        (Tex, tau, xoff_v, and width (gaussian sigma, not FWHM))
        """

        self.modelfunc = modelfunc
        if self.modelfunc is None:
            raise ValueError("Must specify a spectral model function.  See class help for form.")

        if texgrid is None and taugrid is None:
            if path_to_texgrid == '' or path_to_taugrid=='':
                raise IOError("Must specify model grids to use.")
            else:
                self.taugrid = [pyfits.getdata(path_to_taugrid)]
                self.texgrid = [pyfits.getdata(path_to_texgrid)]
                hdr = pyfits.getheader(path_to_taugrid)
                self.yinds,self.xinds = np.indices(self.taugrid[0].shape[1:])
                self.densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
                self.columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
                self.minfreq = (4.8,)
                self.maxfreq = (5.0,)
        elif len(taugrid)==len(texgrid) and hdr is not None:
            self.minfreq,self.maxfreq,self.texgrid = zip(*texgrid)
            self.minfreq,self.maxfreq,self.taugrid = zip(*taugrid)
            self.yinds,self.xinds = np.indices(self.taugrid[0].shape[1:])
            self.densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
            self.columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
        else:
            raise Exception
        
        # Convert X-units to frequency in GHz
        self.xarr = copy.copy(xarr)
        self.xarr.convert_to_unit('Hz', quiet=True)

        #tau = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,taugrid[temperature_gridnumber,:,:])
        #tex = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,texgrid[temperature_gridnumber,:,:])

        if debug:
            import pdb; pdb.set_trace()

    def __call__(self, density=4, column=13, xoff_v=0.0, width=1.0,):
        self.gridval1 = np.interp(density, self.densityarr[0,:], xinds[0,:])
        self.gridval2 = np.interp(column, self.columnarr[:,0], yinds[:,0])
        if np.isnan(gridval1) or np.isnan(gridval2):
            raise ValueError("Invalid column/density")

        if scipyOK:
            tau = [scipy.ndimage.map_coordinates(tg[temperature_gridnumber,:,:],np.array([[self.gridval2],[self.gridval1]]),order=1) for tg in self.taugrid]
            tex = [scipy.ndimage.map_coordinates(tg[temperature_gridnumber,:,:],np.array([[self.gridval2],[self.gridval1]]),order=1) for tg in self.texgrid]
        else:
            raise ImportError("Couldn't import scipy, therefore cannot interpolate")

        if verbose:
            print "density %20.12g column %20.12g: tau %20.12g tex %20.12g" % (density, column, tau, tex)
        if debug:
            import pdb; pdb.set_trace()

        return self.modelfunc(self.xarr,Tex=self.tex,tau=tau,xoff_v=xoff_v,width=width,**kwargs)




try:
    import atpy
    atpyOK = True
except ImportError:
    atpyOK = False

# rewrite this garbage

class write_txt(object):
    def __init__(self, Spectrum):
        self.Spectrum = Spectrum

    def write_data(self, clobber = True):
        """
        Write all fit information to an ASCII file.
        """
        
        fn = "{0}_fit.dat".format(self.Spectrum.fileprefix)
        if not clobber:
            i = 1
            while os.path.exists(fn):
                fn = "{0}_fit({1}).dat".format(self.Spectrum.fileprefix, i)
                i += 1
                
        f = open(fn, 'w')
        
        # Print header
        print >> f, "# Column 1: {0}".format("x-values")
        print >> f, "# Column 2: {0}".format("model spectrum")
        for i, element in enumerate(self.Spectrum.specfit.modelcomponents):
            print >> f, "# Column {0}: model spectrum component {1}".format(i + 3, i + 1)        
        print >> f, "# Column {0}: residuals".format(i + 4)
        print >> f, ""    
                
        components = zip(*self.Spectrum.specfit.modelcomponents)        
        for i, element in enumerate(self.Spectrum.specfit.model):
            line = "{0:10}{1:10}".format(self.Spectrum.xarr[self.Spectrum.specfit.gx1:self.Spectrum.specfit.gx2][i], 
                round(self.Spectrum.specfit.model[i], 5))
            for j, component in enumerate(components[i]): line += "{0:10}".format(round(component, 5))    
            line += "{0:10}".format(round(self.Spectrum.specfit.residuals[i], 5))       
                
            print >> f, line
            
        print >> f, ""
            
        f.close()




"""
==================
Window Interaction
==================

A general module for selecting regions and inputting guesses via the
interactive window.
"""
import numpy
import pyspeckit

class Interactive(object):

    def __init__(self, Spectrum, guesses=None,
            interactive_help_message="Replace this message"):
        """
        Declare interactive variables.

        Must have a parent Spectrum class

        **Must declare button2action and button3action**
        """
        self.Spectrum = Spectrum

        self.interactive_help_message = interactive_help_message
        # includemask should not be a masked array even if data is
        # masked arrays are apparently full of bugs...
        self.includemask = numpy.array( self.Spectrum.data.astype('bool') + True )
        self.xclicks = []
        self.yclicks = []
        self.event_history = []
        self.guesses = guesses

        # Click counters
        self.nclicks_b1 = 0 # button 1
        self.nclicks_b2 = 0 # button 2

        # Temporary storage (for left, right clicking)
        self._xclick1 = None
        self._xclick2 = None

        # Set min/max of range
        self.xmin = 0
        self.xmax = self.Spectrum.xarr.shape[0]

        # Init button 1/2 plots
        self.button1plot = []
        self.button2plot = []

        self.use_window_limits = None

    def event_manager(self, event, debug=False):
        """
        Decide what to do given input (click, keypress, etc.)
        """
        toolbar = self.Spectrum.plotter.figure.canvas.manager.toolbar
        self.event_history.append(event)

        if toolbar.mode == '' and self.Spectrum.plotter.axis in event.canvas.figure.axes:
            if hasattr(event,'button'):
                button = event.button
            elif hasattr(event,'key'):
                button = event.key

            if event.xdata is None or event.ydata is None:
                return

            if debug:
                print "button: ",button," x,y: ",event.xdata,event.ydata

            if button in ('p','P','1',1,'i','a'): # p for... parea?  a for area.  i for include
                # button one is always region selection
                self.selectregion_interactive(event,debug=debug)
            elif button in ('e','x','r','E','X','R'): # e for exclude, x for x-clude, r for remove
                # exclude/delete/remove
                self.selectregion_interactive(event, mark_include=False, debug=debug)
            elif button in ('m','M','2',2): # m for mark
                if debug: print "Button 2 action"
                self.button2action(event,debug=debug)
            elif button in ('d','D','3',3): # d for done
                if debug: print "Button 3 action"
                self.button3action(event,debug=debug)
            elif button in ('?'):
                print self.interactive_help_message
            elif hasattr(self,'Registry') and button in self.Registry.fitkeys:
                fittername = self.Registry.fitkeys[button]
                if fittername in self.Registry.multifitters:
                    self.fitter = self.Registry.multifitters[fittername]
                    self.fittype = fittername
                    print "Selected multi-fitter %s" % fittername
                elif fittername in self.Registry.singlefitters:
                    self.fitter = self.Registry.singlefitters[fittername]
                    self.fittype = fittername
                    print "Selected single-fitter %s" % fittername
                else: 
                    print "ERROR: Did not find fitter %s" % fittername
            if self.Spectrum.plotter.autorefresh: self.Spectrum.plotter.refresh()
        elif debug:
            print "Button press not acknowledged",event


    def selectregion_interactive(self, event, mark_include=True, debug=False, **kwargs):
        """
        select regions for baseline fitting
        """

        xpix = self.Spectrum.xarr.x_to_pix(event.xdata)
        if self.xclicks == []:
            self._firstclick_selection(not mark_include)

        if self.nclicks_b1 == 0:
            self.nclicks_b1 = 1
            self._xclick1 = xpix
            self.xclicks.append(xpix)
            if debug: print "Click 1: clickx=%i xmin=%i, xmax=%i" % (xpix,self.xmin,self.xmax)
        elif self.nclicks_b1 == 1:
            self._xclick2 = xpix
            self.nclicks_b1 = 0
            self.xclicks.append(xpix)
            # force click1 to be left (swap)
            if self._xclick1 > self._xclick2:  
                self._xclick1,self._xclick2 = self._xclick2,self._xclick1

            # ensure that the fit/plot range is at least as large as the click range
            if self.xmin > self._xclick1: self.xmin = self._xclick1
            if self.xmax < self._xclick2: self.xmax = self._xclick2
        
            if mark_include: 
                color = 'g'
                linewidth = 0.5
            else:
                try:
                    color = self.Spectrum.plotter._spectrumplot[0].get_color()
                    linewidth = self.Spectrum.plotter._spectrumplot[0].get_linewidth()
                except IndexError:
                    linewidth = 0.5
                    color = 'k'

            self.button1plot += self.Spectrum.plotter.axis.plot(
                    self.Spectrum.xarr[self._xclick1:self._xclick2],
                    self.Spectrum.data[self._xclick1:self._xclick2]+self.Spectrum.plotter.offset,
                    drawstyle='steps-mid',
                    color=color,
                    linewidth=linewidth)
            self.Spectrum.plotter.refresh()
            self.includemask[self._xclick1:self._xclick2] = mark_include
            if debug: print "Click 2: clickx=%i xmin=%i, xmax=%i" % (xpix,self.xmin,self.xmax)

        self._update_xminmax()

    def highlight_fitregion(self,  drawstyle='steps-mid', color='g',
            clear_highlights=True, **kwargs):
        """
        Re-highlight the fitted region

        kwargs are passed to `matplotlib.plot`
        """
        if clear_highlights: self.clear_highlights()
        bad = self.Spectrum.data*0
        bad[True-self.includemask] = numpy.nan
        self.button1plot += self.Spectrum.plotter.axis.plot(
                self.Spectrum.xarr,
                self.Spectrum.data+self.Spectrum.plotter.offset+bad,
                drawstyle=drawstyle, color=color, 
                **kwargs)
        self.Spectrum.plotter.refresh()

    def _firstclick_selection(self, include_all=False):
        """
        Initialize the include/exclude mask
        """

        self.Spectrum.plotter.axis.set_autoscale_on(False)
        if include_all:
            # default to including everything
            self.includemask = numpy.array(self.Spectrum.data, dtype='bool') + True
        else:
            # default to including nothing
            self.includemask = numpy.array(self.Spectrum.data, dtype='bool') * False

    def guesspeakwidth(self,event,debug=False):
        """
        Interactively guess the peak height and width from user input

        Width is assumed to be half-width-half-max
        """
        if self.nclicks_b2 == 0:
            self.firstclick_guess()
        if self.nclicks_b2 % 2 == 0:
            if self.Spectrum.baseline.subtracted:
                peakguess = event.ydata
            else:
                peakguess = event.ydata - self.Spectrum.baseline.basespec[self.Spectrum.xarr.x_to_pix(event.xdata)]
            self.guesses += [peakguess,event.xdata,1]
            self.npeaks += 1
            self.nclicks_b2 += 1
            if debug: print "Peak %i click %i at x,y %g,%g" % (self.npeaks,self.nclicks_b2,event.xdata,event.ydata)
            self.button2plot += [self.Spectrum.plotter.axis.scatter(event.xdata,event.ydata,marker='x',c='r')]
            #self.Spectrum.plotter.refresh() #plot(**self.Spectrum.plotter.plotkwargs)
        elif self.nclicks_b2 % 2 == 1:
            self.guesses[-1] = abs(event.xdata-self.guesses[-2]) / numpy.sqrt(2*numpy.log(2))
            self.nclicks_b2 += 1
            if debug: print "Width %i click %i at x,y %g,%g" % (self.npeaks,self.nclicks_b2,event.xdata,event.ydata)
            self.button2plot += self.Spectrum.plotter.axis.plot([event.xdata,
                2*self.guesses[-2]-event.xdata],[event.ydata]*2,
                color='r')
            #self.Spectrum.plotter.refresh() #plot(**self.Spectrum.plotter.plotkwargs)
            if self.nclicks_b2 / 2 > self.npeaks:
                print "There have been %i middle-clicks but there are only %i gaussians" % (self.nclicks_b2,self.npeaks)
                self.npeaks += 1

    def firstclick_guess(self):
        """
        Initialize self.guesses
        """
        self.Spectrum.plotter.axis.set_autoscale_on(False)
        if self.guesses is None:
            self.guesses = []
        elif len(self.guesses) > 0:
            for ii in xrange(len(guesses)): 
                self.guesses.pop()

    def clear_all_connections(self, debug=False):
        """
        Prevent overlapping interactive sessions
        """
        # this is really ugly, but needs to be done in order to prevent multiple overlapping calls...
        cids_to_remove = []
        if not hasattr(self.Spectrum.plotter.figure,'canvas'):
            # just quit out; saves a tab...
            if debug: print "Didn't find a canvas, quitting."
            return
        for eventtype in ('button_press_event','key_press_event'):
            for key,val in self.Spectrum.plotter.figure.canvas.callbacks.callbacks[eventtype].iteritems():
                if "event_manager" in val.func.__name__:
                    cids_to_remove.append(key)
                    if debug: print "Removing CID #%i with attached function %s" % (key,val.func.__name__)
        for cid in cids_to_remove:
            self.Spectrum.plotter.figure.canvas.mpl_disconnect(cid)

        self.Spectrum.plotter._reconnect_matplotlib_keys()


    def start_interactive(self, debug=False, LoudDebug=False, print_message=True, clear_all_connections=True, **kwargs):
        """

        """
        if print_message: print self.interactive_help_message
        if clear_all_connections: 
            self.clear_all_connections()
            self.Spectrum.plotter._disconnect_matplotlib_keys()
        key_manager = lambda(x): self.event_manager(x, debug=debug, **kwargs)
        click_manager = lambda(x): self.event_manager(x, debug=debug, **kwargs)
        key_manager.__name__ = "event_manager"
        click_manager.__name__ = "event_manager"
        self.click    = self.Spectrum.plotter.axis.figure.canvas.mpl_connect('button_press_event',click_manager)
        self.keyclick = self.Spectrum.plotter.axis.figure.canvas.mpl_connect('key_press_event',key_manager)
        self._callbacks = self.Spectrum.plotter.figure.canvas.callbacks.callbacks
        self._check_connections()

    def _check_connections(self):
        """
        Make sure the interactive session acepts user input
        """
        # check for connections
        OKclick = False
        OKkey   = False
        for cb in self._callbacks.values():
            if self.click in cb.keys():
                OKclick = True
            if self.keyclick in cb.keys():
                OKkey = True
        if self.keyclick == self.click:
            OKkey = False
        if not OKkey:
            print "Interactive session failed to connect keyboard.  Key presses will not be accepted."
        if not OKclick:
            print "Interactive session failed to connect mouse.  Mouse clicks will not be accepted."


    def clear_highlights(self):
        for p in self.button1plot:
            p.set_visible(False)
            if p in self.Spectrum.plotter.axis.lines: self.Spectrum.plotter.axis.lines.remove(p)
        self.button1plot=[] # I should be able to just remove from the list... but it breaks the loop...
        self.Spectrum.plotter.refresh()

    def selectregion(self, xmin=None, xmax=None, xtype='wcs', highlight=False,
            fit_plotted_area=True, reset=False, verbose=False, debug=False,
            use_window_limits=None, exclude=None, **kwargs):
        """
        Pick a fitting region in either WCS units or pixel units

        *xmin / xmax* [ float ]
            The min/max X values to use in X-axis units (or pixel units if xtype is set).
            TAKES PRECEDENCE ALL OTHER BOOLEAN OPTIONS

        *xtype* [ string ]
            A string specifying the xtype that xmin/xmax are specified in.  It can be either
            'wcs' or any valid xtype from :class:`pyspeckit.spectrum.units`

        *reset* [ bool ]
            Reset the selected region to the full spectrum?  Only takes effect
            if xmin and xmax are not (both) specified.   
            TAKES PRECEDENCE ALL SUBSEQUENT BOOLEAN OPTIONS

        Now things get a little complicated...
        *fit_plotted_area* [ bool ]
            Use the plot limits *as specified in :class:`pyspeckit.spectrum.plotters`*?
            Note that this is not necessarily the same as the window plot limits!

        *use_window_limits* [ bool ] 
            Use the plot limits *as displayed*.  Defaults to
            :attr:`pyspeckit.spectrum.interactive.use_window_limits`.
            Overwrites xmin,xmax set by plotter
            
        exclude: {list of length 2n,'interactive', None}
            * interactive: start an interactive session to select the
              include/exclude regions
            * list: parsed as a series of (startpoint, endpoint) in the
              spectrum's X-axis units.  Will exclude the regions between
              startpoint and endpoint
            * None: No exclusion
        """
        if debug: print "selectregion kwargs: ",kwargs," use_window_limits: ",use_window_limits," reset: ",reset," xmin: ",xmin, " xmax: ",xmax
        if xmin is not None and xmax is not None:
            if verbose or debug: print "Setting xmin,xmax from keywords %g,%g" % (xmin,xmax)
            if xtype.lower() in ('wcs',) or xtype in pyspeckit.spectrum.units.xtype_dict:
                self.xmin = numpy.floor(self.Spectrum.xarr.x_to_pix(xmin))
                self.xmax = numpy.ceil(self.Spectrum.xarr.x_to_pix(xmax))
            else:
                self.xmin = xmin
                self.xmax = xmax
        elif reset:
            if verbose or debug: print "Resetting xmin/xmax to full limits of data"
            self.xmin = 0
            self.xmax = self.Spectrum.data.shape[0]
            self.includemask[self.xmin:self.xmax] = True
            #raise ValueError("Need to input xmin and xmax, or have them set by plotter, for selectregion.")
        elif self.Spectrum.plotter.xmin is not None and self.Spectrum.plotter.xmax is not None and fit_plotted_area:
            if use_window_limits or (use_window_limits is None and self.use_window_limits):
                if debug: print "Resetting plotter xmin,xmax and ymin,ymax to the currently visible region"
                self.Spectrum.plotter.set_limits_from_visible_window(debug=debug)
            self.xmin = numpy.floor(self.Spectrum.xarr.x_to_pix(self.Spectrum.plotter.xmin))
            self.xmax = numpy.ceil(self.Spectrum.xarr.x_to_pix(self.Spectrum.plotter.xmax))
            if self.xmin>self.xmax: 
                self.xmin,self.xmax = self.xmax,self.xmin
            if debug: print "Including all plotted area (as defined by [plotter.xmin,plotter.xmax]) for fit"
            if debug: print "Including self.xmin:self.xmax = %f:%f (and excluding the rest)" % (self.xmin,self.xmax)
            self.includemask[:self.xmin] = False
            self.includemask[self.xmax:] = False
            self.includemask[self.xmin:self.xmax] = True
        else:
            if verbose: print "Left region selection unchanged.  xminpix, xmaxpix: %i,%i" % (self.xmin,self.xmax)
        
        if self.xmin == self.xmax:
            # Reset if there is no fitting region
            self.xmin = 0
            self.xmax = self.Spectrum.data.shape[0]
            if debug: print "Reset to full range because the endpoints were equal"
        elif self.xmin>self.xmax: 
            # Swap endpoints if the axis has a negative delta-X
            self.xmin,self.xmax = self.xmax,self.xmin
            if debug: print "Swapped endpoints because the left end was greater than the right"

        self.includemask[:self.xmin] = False
        self.includemask[self.xmax:] = False

        # Exclude keyword-specified excludes.  Assumes exclusion in current X array units
        if exclude is not None and len(exclude) % 2 == 0:
            for x1,x2 in zip(exclude[::2],exclude[1::2]):
                x1 = self.Spectrum.xarr.x_to_pix(x1)
                x2 = self.Spectrum.xarr.x_to_pix(x2)
                self.includemask[x1:x2] = False

        if highlight:
            self.highlight_fitregion()
        
        self._update_xminmax()

    def _update_xminmax(self):
        try:
            whinclude = numpy.where(self.includemask)
            self.xmin = whinclude[0][0]
            self.xmax = whinclude[0][-1]
        except IndexError:
            pass




"""
===============
Gaussian Fitter
===============

The simplest and most useful model.

Until 12/23/2011, gaussian fitting used the complicated and somewhat bloated
gaussfitter.py code.  Now, this is a great example of how to make your own
model!  Just make a function like gaussian and plug it into the SpectralModel
class.

"""
import model
import numpy 

def gaussian(x,A,dx,w, return_components=False):
    """
    Returns a 1-dimensional gaussian of form
    H+A*numpy.exp(-(x-dx)**2/(2*w**2))
    
    [height,amplitude,center,width]

    return_components does nothing but is required by all fitters
    
    """
    x = numpy.array(x) # make sure xarr is no longer a spectroscopic axis
    return A*numpy.exp(-(x-dx)**2/(2.0*w**2))

def gaussian_fitter(multisingle='multi'):
    """
    Generator for Gaussian fitter class
    """

    myclass =  model.SpectralModel(gaussian, 3,
            parnames=['amplitude','shift','width'], 
            parlimited=[(False,False),(False,False),(True,False)], 
            parlimits=[(0,0), (0,0), (0,0)],
            shortvarnames=('A',r'\Delta x',r'\sigma'),
            multisingle=multisingle,
            )
    myclass.__name__ = "gaussian"
    
    return myclass



"""
Ammonia inversion transition TKIN fitter translated from Erik Rosolowsky's
http://svn.ok.ubc.ca/svn/signals/nh3fit/
"""
import numpy as np
from pyspeckit.mpfit import mpfit
from .. import units
import matplotlib.cbook as mpcb

line_names = ['oneone','twotwo','threethree']
line_names = ['oneone_f10','oneone_f01','oneone_f22','oneone_f21','oneone_f12','oneone_f11']

# http://articles.adsabs.harvard.edu/abs/1971ApJ...169..429T has the most accurate freqs
freq_dict = { 
    'oneone':     4.82965996e9,
    'twotwo':     14.48848e9,
    'threethree': 28.97480e9,
    }
relative_strength_theory={
        'oneone_f10': 4,
        'oneone_f01': 4,
        'oneone_f22':15,
        'oneone_f21': 5,
        'oneone_f12': 5,
        'oneone_f11': 3,
        'twotwo_f11':1,
        'twotwo_f12':1,
        'twotwo_f21':1,
        'twotwo_f32':1,
        'twotwo_f33':1,
        'twotwo_f22':1,
        'twotwo_f23':1,
        'threethree_f22':1,
        'threethree_f44':1,
        'threethree_f33':1,
        }
hf_freq_dict={
        'oneone_f10':4.82965996e9 - 18.53e3,
        'oneone_f01':4.82965996e9 - 1.34e3,
        'oneone_f22':4.82965996e9 - 0.35e3,
        'oneone_f21':4.82965996e9 + 4.05e3,
        'oneone_f12':4.82965996e9 + 6.48e3,
        'oneone_f11':4.82965996e9 + 11.08e3,
        'twotwo_f11':14.48846e9,
        'twotwo_f12':14.48847e9,
        'twotwo_f21':14.48848e9,
        'twotwo_f32':14.48848e9,
        'twotwo_f33':14.48848e9,
        'twotwo_f22':14.48849e9,
        'twotwo_f23':14.48849e9,
        'threethree_f22':28.97478e9,
        'threethree_f44':28.97480e9,
        'threethree_f33':28.97481e9,
        }
freq_dict.update(hf_freq_dict)
aval_dict = {
    'oneone':     10**-8.44801,  #64*!pi**4/(3*h*c**3)*nu11**3*mu0**2*(1/2.)
    'twotwo':     10**-7.49373,  #64*!pi**4/(3*h*c**3)*nu22**3*mu0**2*(2/3.)
    'threethree': 10**-6.89179,  #64*!pi**4/(3*h*c**3)*nu33**3*mu0**2*(3/4.)
    }
hf_aval_dict={
        'oneone_f10':10**-8.92509,
        'oneone_f01':10**-8.44797,
        'oneone_f22':10**-8.57294,
        'oneone_f21':10**-9.05004,
        'oneone_f12':10**-8.82819,
        'oneone_f11':10**-9.05009,
        'twotwo_f11':10**-7.61876,
        'twotwo_f12':10**-8.09586,
        'twotwo_f21':10**-8.31771,
        'twotwo_f32':10**-8.44804,
        'twotwo_f33':10**-7.54494,
        'twotwo_f22':10**-7.65221,
        'twotwo_f23':10**-8.30191,
        'threethree_f22':10**-6.94294,
        'threethree_f44':10**-6.91981,
        'threethree_f33':10**-6.96736,
        }
ortho_dict = {
    'oneone':     False,
    'twotwo':     False,
    'threethree': False,
    }
n_ortho = np.arange(0,28,3) # 0..3..27
n_para = np.array([x for x in range(28) if x % 3 != 0])

voff_lines_dict = {
        'oneone': [(hf_freq_dict[f]-freq_dict['oneone'])/freq_dict['oneone']*units.speedoflight_ms for f in hf_freq_dict.keys() if "oneone" in f],
        'twotwo': [(hf_freq_dict[f]-freq_dict['twotwo'])/freq_dict['twotwo']*units.speedoflight_ms for f in hf_freq_dict.keys() if "twotwo" in f],
        'threethree': [(hf_freq_dict[f]-freq_dict['threethree'])/freq_dict['threethree']*units.speedoflight_ms for f in hf_freq_dict.keys() if "threethree" in f],
        }
voff_lines_dict={
        'oneone_f10': 18.53e3/4.82965996e9 * units.speedoflight_ms / 1000.0,
        'oneone_f01': 1.34e3 /4.82965996e9 * units.speedoflight_ms / 1000.0,
        'oneone_f22': 0.35e3 /4.82965996e9 * units.speedoflight_ms / 1000.0,
        'oneone_f21': 4.05e3 /4.82965996e9 * units.speedoflight_ms / 1000.0,
        'oneone_f12': 6.48e3 /4.82965996e9 * units.speedoflight_ms / 1000.0,
        'oneone_f11': 11.08e3/4.82965996e9 * units.speedoflight_ms / 1000.0,
        'twotwo_f11':14.48846e9,
        'twotwo_f12':14.48847e9,
        'twotwo_f21':14.48848e9,
        'twotwo_f32':14.48848e9,
        'twotwo_f33':14.48848e9,
        'twotwo_f22':14.48849e9,
        'twotwo_f23':14.48849e9,
        'threethree_f22':28.97478e9,
        'threethree_f44':28.97480e9,
        'threethree_f33':28.97481e9,
        }



class formaldehyde_model(object):

    def __init__(self):
        self.npeaks = 1
        self.npars = 6
        pass

    def formaldehyde(self, xarr, xunits='GHz', amp=1.0, width=1.0,
            xoff_v=0.0, line='oneone'):
        """
        Generate a model Formaldehyde spectrum based on input temperatures, column, and
        gaussian parameters


        (not implemented) if tau11 is specified, Ntot is ignored
        """

        # Convert X-units to frequency in GHz
        if xunits in units.frequency_dict:
            xarr = np.copy(xarr) * units.frequency_dict[xunits] / units.frequency_dict['GHz']
        elif xunits in units.velocity_dict:
            if line in freq_dict:
                xarr = (freq_dict[line] - (np.copy(xarr) * 
                        (units.velocity_dict[xunits] / units.velocity_dict['m/s'] / units.speedoflight_ms) *
                        freq_dict[line]) ) / units.frequency_dict['GHz']
            else:
                raise Exception("Xunits is velocity-type (%s) but line %s is not in the list." % (xunits,line))
        else:
            raise Exception("xunits not recognized: %s" % (xunits))

        ckms = 2.99792458e5
        ccms = ckms*1e5
        g1 = 1                
        g2 = 1                
        h = 6.6260693e-27     
        kb = 1.3806505e-16     

        runspec = np.zeros(len(xarr))

        for linename in line_names:
            voff_lines = np.array(voff_lines_dict[linename])
      
            lines = (1-voff_lines/ckms)*freq_dict[linename]
            nuwidth = np.abs(width/ckms*lines)
            nuoff = xoff_v/ckms*lines
      
            # strength array
            runspec += (1-relative_strength_theory[linename]*amp*\
                    np.exp(-(xarr+nuoff-freq_dict[linename])**2/(2*nuwidth**2)))
      
        return runspec

    def n_formaldehyde(self, pars=None, fittau=False, **kwargs):
        """
        Returns a function that sums over N ammonia line profiles, where N is the length of
        tkin,tex,Ntot,width,xoff_v,fortho *OR* N = len(pars) / 6

        The background "height" is assumed to be zero (you must "baseline" your
        spectrum before fitting)

        pars  - a list with len(pars) = 6n, assuming tkin,tex,Ntot,width,xoff_v,fortho repeated
        """
        if len(pars) % 6 == 0:
            tkin = [pars[ii] for ii in xrange(0,len(pars),6)]
            tex = [pars[ii] for ii in xrange(1,len(pars),6)]
            Ntot = [pars[ii] for ii in xrange(2,len(pars),6)]
            width = [pars[ii] for ii in xrange(3,len(pars),6)]
            xoff_v = [pars[ii] for ii in xrange(4,len(pars),6)]
            fortho = [pars[ii] for ii in xrange(5,len(pars),6)]
        elif not(len(tkin) == len(tex) == len(Ntot) == len(xoff_v) == len(width) == len(fortho)):
            raise ValueError("Wrong array lengths!")

        modelkwargs = kwargs.copy()
        def L(x):
            v = np.zeros(len(x))
            for i in range(len(tkin)):
                modelkwargs.update({'tkin':tkin[i], 'tex':tex[i],
                        'width':width[i], 'xoff_v':xoff_v[i],
                        'fortho':fortho[i]})
                if fittau:
                    modelkwargs.update({'tau11':Ntot[i]})
                else:
                    modelkwargs.update({'Ntot':Ntot[i]})
                v += self.ammonia(x,**modelkwargs)
            return v
        return L

    def multinh3fit(self, xax, data, npeaks=1, err=None, params=[20,20,1e10,1.0,0.0,0.5],
            fixed=[False,False,False,False,False,False],
            limitedmin=[True,True,True,True,False,True],
            limitedmax=[False,False,False,False,False,True], minpars=[2.73,2.73,0,0,0,0],
            maxpars=[0,0,0,0,0,1], quiet=True, shh=True, veryverbose=False, **kwargs):
        """
        Fit multiple nh3 profiles

        Inputs:
           xax - x axis
           data - y axis
           npeaks - How many nh3 profiles to fit?  Default 1 (this could supersede onedgaussfit)
           err - error corresponding to data

         These parameters need to have length = 6*npeaks.  If npeaks > 1 and length = 6, they will
         be replicated npeaks times, otherwise they will be reset to defaults:
           params - Fit parameters: [amplitude, offset, Gfwhm, Lfwhm] * npeaks
                  If len(params) % 6 == 0, npeaks will be set to len(params) / 6
           fixed - Is parameter fixed?
           limitedmin/minpars - set lower limits on each parameter (default: width>0, Tex and Tkin > Tcmb)
           limitedmax/maxpars - set upper limits on each parameter

           quiet - should MPFIT output each iteration?
           shh - output final parameters?

        Returns:
           Fit parameters
           Model
           Fit errors
           chi2
        """

        self.npars = 6

        if len(params) != npeaks and (len(params) / self.npars) > npeaks:
            npeaks = len(params) / self.npars 
        self.npeaks = npeaks

        if isinstance(params,np.ndarray): params=params.tolist()

        # make sure all various things are the right length; if they're not, fix them using the defaults
        for parlist in (params,fixed,limitedmin,limitedmax,minpars,maxpars):
            if len(parlist) != self.npars*self.npeaks:
                # if you leave the defaults, or enter something that can be multiplied by 3 to get to the
                # right number of gaussians, it will just replicate
                if len(parlist) == self.npars: 
                    parlist *= npeaks 
                elif parlist==params:
                    parlist[:] = [20,20,1e10,1.0,0.0,0.5] * npeaks
                elif parlist==fixed:
                    parlist[:] = [False,False,False,False,False,False] * npeaks
                elif parlist==limitedmax:
                    parlist[:] = [False,False,False,False,False,True] * npeaks
                elif parlist==limitedmin:
                    parlist[:] = [True,True,True,True,False,True] * npeaks
                elif parlist==minpars:
                    parlist[:] = [2.73,0,0,0,0,0] * npeaks
                elif parlist==maxpars:
                    parlist[:] = [0,0,0,0,0,1] * npeaks

        def mpfitfun(x,y,err):
            if err is None:
                def f(p,fjac=None): return [0,(y-self.n_ammonia(pars=p, **kwargs)(x))]
            else:
                def f(p,fjac=None): return [0,(y-self.n_ammonia(pars=p, **kwargs)(x))/err]
            return f

        parnames = {0:"TKIN",1:"TEX",2:"NTOT",3:"WIDTH",4:"XOFF_V",5:"FORTHO"}

        parinfo = [ {'n':ii, 'value':params[ii],
            'limits':[minpars[ii],maxpars[ii]],
            'limited':[limitedmin[ii],limitedmax[ii]], 'fixed':fixed[ii],
            'parname':parnames[ii%self.npars]+str(ii/self.npars), 
            'mpmaxstep':0,'error':ii} 
            for ii in xrange(len(params)) ]
        parinfo[0]['mpmaxstep'] = 1.0
        parinfo[1]['mpmaxstep'] = 1.0

        if veryverbose:
            print "GUESSES: "
            print "\n".join(["%s: %s" % (p['parname'],p['value']) for p in parinfo])

        mp = mpfit(mpfitfun(xax,data,err),parinfo=parinfo,quiet=quiet)
        mpp = mp.params
        if mp.perror is not None: mpperr = mp.perror
        else: mpperr = mpp*0
        chi2 = mp.fnorm

        if mp.status == 0:
            raise Exception(mp.errmsg)

        if not shh:
            print "Fit message: ",mp.errmsg
            print "Final fit values: "
            for i,p in enumerate(mpp):
                parinfo[i]['value'] = p
                print parinfo[i]['parname'],p," +/- ",mpperr[i]
            print "Chi2: ",mp.fnorm," Reduced Chi2: ",mp.fnorm/len(data)," DOF:",len(data)-len(mpp)

        if mpp[1] > mpp[0]: mpp[1] = mpp[0]  # force Tex>Tkin to Tex=Tkin (already done in n_ammonia)
        self.mp = mp
        self.mpp = mpp
        self.mpperr = mpperr
        self.model = self.n_ammonia(pars=mpp,**kwargs)(xax)
        return mpp,self.n_ammonia(pars=mpp,**kwargs)(xax),mpperr,chi2

    __call__ = multinh3fit

    def moments(self, Xax, data, negamp=None, veryverbose=False,  **kwargs):
        """
        Returns a very simple and likely incorrect guess
        """

        # TKIN, TEX, NTOT, width, center, ortho fraction
        return [20,10, 1e15, 1.0, 0.0, 1.0]

    def annotations(self):
        label_list = [ (
                "$T_K(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[0+jj*self.npars],self.mpperr[0+jj*self.npars]),
                "$T_{ex}(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[1+jj*self.npars],self.mpperr[1+jj*self.npars]),
                "$N$(%i)=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[2+jj*self.npars],self.mpperr[2+jj*self.npars]),
                "$\\sigma(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[3+jj*self.npars],self.mpperr[3+jj*self.npars]),
                "$v(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[4+jj*self.npars],self.mpperr[4+jj*self.npars]),
                "$F_o(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[5+jj*self.npars],self.mpperr[5+jj*self.npars])
                          ) for jj in range(self.npeaks)]
        labels = tuple(mpcb.flatten(label_list))
        return labels



"""
A flexible Spectral Model class.  Allows you to add and subtract, fix and unfix functions, etc.

Goal is to allow adding/subtracting/multiplying components...
(like these: https://astrophysics.gsfc.nasa.gov/XSPECwiki/Basic_single_dataset)
"""

import numpy as np


class ModelComponent(object):
    def __init__(self, function, name, parameters=None, fixed=False):
        """
        Each Model object consists of a function...
        """
        self.function = function
        self.name = name
        self.parameters = parameters
        self.fixed = fixed

    def __call__(self, xaxis, parameters):
        """
        A model is defined over an X-axis.  It can do anything here, but must
        return an array of length len(xaxis).
        """

        return self.function(xaxis, parameters)

        return sum_of_model_functions

class AdditiveModelComponent(ModelComponent):
    pass
    
class MultiplicativeModelComponent(ModelComponent):
    pass
    

class CompositeModel(ModelComponent):
    """
    A model will consist of a sum of its children?

    """
    def __init__(self):
        """
        Generic model...
        """
        self.additive_components = {}
        self.multiplicative_components = {}
        self.parameters = {}
        self.npars = 0

    def __call__(self, xaxis):
        """
        Return sum/
        """

        model = np.zeros(xaxis.shape)

        for name,func in additive_components.iteritems():
            #model += func(xaxis, ??parameters??)
            pass

    def __add__(self, other):
        """
        Models can be added together
        """
        self.additive_components[other.name] = other
        self.parameters[other.name] = other.parameters
        for ii in xrange(len(other.parameters)):
            self.parameters[self.npars+ii] = other.parameters[ii]
        self.npars += len(other.parameters)

        return self

    def __sub__(self, other):
        """
        Models can be removed
        """
        try:
            self.additive_components.remove(other)
        except ValueError:
            print "Can only remove components that have been included in the model"

        return self

    def __mul__(self, other):
        """
        Models can be added multiplicatively too
        """
        self.multiplicative_components[other.name] = other
        self.parameters[other.name] = other.parameters
        for ii in xrange(len(other.parameters)):
            self.parameters[self.npars+ii] = other.parameters[ii]
        self.npars += len(other.parameters)
        return self

    def __div__(self, other):
        """
        And models can be divided out
        """
        try:
            self.multiplicative_components.remove(other)
        except ValueError:
            print "Can only remove components that have been included in the model"

        return self


if __name__ == "__main__":
    # unit test

    test_xaxis = np.linspace(-5,5,100)
    test_spectrum = (np.exp(-(test_xaxis-1.2)**2/2.)*1.5 +
                    np.exp(-(test_xaxis-1.1)**2/(2.*1.5**2))*1.0 +
                    (6.0+test_xaxis*1.5) )

    import pyspeckit.spectrum.models as models

    # not implemented yet, but this is how it should work
    MyModel = CompositeModel()
    MyModel += Model(models.inherited_gaussfitter.gaussian,'gaussian1',params=[1,0,0])
    MyModel += Model(models.inherited_gaussfitter.gaussian,'gaussian2',params=[1,0,0])
    MyModel += Model(lambda x,m,b: m*x+b,'line1',params=[1,0])

    MyModel.fit(test_xaxis, data)
    MyModel.gaussian1.fixed[:] = True



"""
Logger for the spectroscopic toolkit packge

Goal: Use decorators to log each command in full

Author: Adam Ginsburg
Created: 03/17/2011
"""
import os
import time

class Logger(object):
    """
    Logger object.  Should be initiated on import.
    """

    def __init__(self, filename):
        """
        Open a log file for writing
        """

        newfilename = filename
        ii=0
        while os.path.exists(filename):
            newfilename = filename+".%3i" % ii
            ii += 1
        if newfilename != filename:
            os.rename(filename,newfilename)

        self.outfile = open(filename,'w')
        print >>self.outfile,"Began logging at %s" % (time.strftime("%M/%d/%Y %H:%M:%S",time.localtime()))

    def __call__(self, function):
        """
        """
        def wrapper(*args, **kwargs):
            modname = function.__module__
            fname = function.__name__
            print >>self.outfile,"%s.%s" % (modname,fname) +\
                    "("+"".join([a.__name__ for a in args]) + \
                    "".join(["%s=%s" % (k,v) for (k,v) in kwargs])+ ")"
            return function
        return wrapper

    def close(self):
        self.outfile.close()



"""
==========================
PySpecKit config system
==========================

Create decorator used to modify inputs to various __call__ methods
to reflect preferences we set in ~/.pyspeckit/config.

To see what values exist in config file, do:
    from config import mycfg
    mycfg.keys()
    
To decorate a new __call__ method, do:
    from config import ConfigDescriptor as cfgdec
    
    @cfgdec
    def __call__(self, **kwargs):
        pass    # do something!

.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
.. moduleauthor:: Jordan Mirocha <mirochaj@gmail.com>
"""

import os, inspect

class dotdictify(dict):
    """
    Class grabbed from
    http://stackoverflow.com/questions/3031219/python-recursively-access-dict-via-attributes-as-well-as-index-access
    to allow "dot" access to dictionary keys
    """
    marker = object()
    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError, 'expected dict'

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, dotdictify):
            value = dotdictify(value)
        dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        found = self.get(key, dotdictify.marker)
        if found is dotdictify.marker:
            found = dotdictify()
            dict.__setitem__(self, key, found)
        return found

    __setattr__ = __setitem__
    __getattr__ = __getitem__


cfgDefaults = dotdictify(dict(
    color = 'k',
    composite_fit_color = 'r',
    component_fit_color = 'blue',
    baseline_fit_color = 'orange',
    lw = 0.5,
    composite_lw = 0.75,
    component_lw = 0.75,
    show_components = False,
    autoannotate = True,
    interactive = False,
    autorefresh = True,
    silent = True,
    debug = False,
    WARN = True,
    ))

class ConfigParser:
    def __init__(self, fn = None):
        """
        Initialize cfg dictionary.
        """
        if fn is not None:
            f = open(fn)
            return_dict = cfgDefaults
            for line in f:
                if not line.strip(): continue
                thisline = line.split()
            
                if thisline[0][0] == '#': continue
                
                if thisline[2] in ['True', 1]: return_dict[thisline[0]] = True
                elif thisline[2] in ['False', 0]: return_dict[thisline[0]] = False
                elif thisline[2] == 'None': return_dict[thisline[0]] = None
                elif thisline[2].isalpha(): return_dict[thisline[0]] = str(thisline[2])
                else: return_dict[thisline[0]] = float(thisline[2])
    	            	        
    	    self.cfg = dotdictify(return_dict)
    	else: 
    	    self.cfg = dotdictify(cfgDefaults)
            	    	
__fn = os.path.expanduser("~/.pyspeckit/config")
if os.path.exists(__fn): 
    mycfg = dotdictify(ConfigParser(__fn).cfg)
else:
    mycfg = dotdictify(ConfigParser().cfg)
   
def ConfigDescriptor(f):    
                                
    def decorator(self, *args, **kwargs):
        """
        This is our decorator function, used to modify the inputs of __call__
        methods to reflect preferences we set in config file.
                
        Notes:
        inspect.getargspec will tell us the names of all arguments and their default values.
        Later we'll have to be more careful - all_defs only makes entries for arguments that actually
        have default values          
        """  

        all_args, all_vars, all_keys, all_defs = inspect.getargspec(f)                
        all_args.pop(0) # pop self
        
        # Construct dictionary containing all of f's keyword arguments                                   
        argsdefs = {}
        for i, arg in enumerate(all_args):
            argsdefs[arg] = all_defs[i]
                                                                   
        # Include these in our new_kwargs dictionary                                                                                                     
        new_kwargs = argsdefs
        
        # Read in config file and replace keyword arguments that have been defined in it
        for arg in new_kwargs:
            if arg in mycfg: new_kwargs[arg] = mycfg[arg]
                
        # If we've changed anything on call, reflect this in new_kwargs
        for arg in kwargs:
            new_kwargs[arg] = kwargs[arg]       
                                                                                                                                       
        f(self, *args, **new_kwargs)
            
    # documentation should be passed on, else sphinx doesn't work and the user can't access the docs
    decorator.__doc__ = f.__doc__
    decorator.__defaults__ = f.__defaults__
    decorator.__repr__ = f.__repr__
    return decorator      




import numpy as np
import pyspeckit
import pyfits
from pyspeckit.spectrum import models
from pyspeckit.wrappers import fith2co

path_to_data = "/Users/adam/work/h2co/radex/troscompt_grid_March2012"

texgrid1 = pyfits.getdata(path_to_data+'/1-1_2-2_T=5to55_lvg_troscompt_100square_opgrid_tex1.fits')
taugrid1 = pyfits.getdata(path_to_data+'/1-1_2-2_T=5to55_lvg_troscompt_100square_opgrid_tau1.fits')
texgrid2 = pyfits.getdata(path_to_data+'/1-1_2-2_T=5to55_lvg_troscompt_100square_opgrid_tex2.fits')
taugrid2 = pyfits.getdata(path_to_data+'/1-1_2-2_T=5to55_lvg_troscompt_100square_opgrid_tau2.fits')
hdr    = pyfits.getheader(path_to_data+'/1-1_2-2_T=5to55_lvg_troscompt_100square_opgrid_tau2.fits')
# # this deserves a lot of explanation:
# # models.formaldehyde.formaldehyde_radex is the MODEL that we are going to fit
# # models.model.SpectralModel is a wrapper to deal with parinfo, multiple peaks,
# # and annotations
# # all of the parameters after the first are passed to the model function
formaldehyde_radex_fitter = models.model.SpectralModel(
        models.formaldehyde.formaldehyde_radex_orthopara_temp, 6,
        parnames=['density','column','orthopara','temperature','center','width'], 
        parvalues=[4,12,1.0,15.0,0,1],
        parlimited=[(True,True), (True,True), (True,True), (True,True), (False,False), (True,False)], 
        parlimits=[(1,8), (11,16), (-3,np.log10(3.0)), (5,55), (0,0), (0,0)],
        parsteps=[0.01,0.01,0,0,0,0],
        fitunits='Hz',
        texgrid=((4,5,texgrid1),(14,15,texgrid2)),
        taugrid=((4,5,taugrid1),(14,15,taugrid2)),
        hdr=hdr,
        shortvarnames=("n","N",'OP','T',"v","\\sigma"),
        )



"""
===========================
Units and SpectroscopicAxes
===========================

Unit parsing and conversion tool.  
The SpectroscopicAxis class is meant to deal with unit conversion internally

Open Questions: Are there other FITS-valid projection types, unit types, etc.
    that should be included?
    What about for other fields (e.g., wavenumber?)
"""
#.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
#Affiliation: University of Colorado at Boulder
#Created on March 9th, 2011


import numpy as np
import warnings

# declare a case-insensitive dict class to return case-insensitive versions of each dictionary...
# this is just a shortcut so that units can be specified as, e.g., Hz, hz, HZ, hZ, etc.  3 of those 4 are "legitimate".  
# the templates for this code were grabbed from a few StackOverflow.com threads
class CaseInsensitiveDict(dict):
    def __init__(self, inputdict=None):
        if inputdict:
            # Doesn't do keyword args
            if isinstance(inputdict, dict):
                self.CaseDict = inputdict
                for k,v in inputdict.items():
                    if hasattr(k,'lower'):
                        dict.__setitem__(self, k.lower(), v)
            else:
                for k,v in inputdict:
                    if hasattr(k,'lower'):
                        dict.__setitem__(self, k.lower(), v)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.lower())

    def __setitem__(self, key, value):
        dict.__setitem__(self, key.lower(), value)

    def __contains__(self, key):
        if hasattr(key,'lower'):
            return dict.__contains__(self, key.lower())
        else: # e.g., None will go here
            return dict.__contains__(self, key)

    def has_key(self, key):
        """ This is deprecated, but we're keeping it around """
        return dict.has_key(self, key.lower())

    def get(self, key, def_val=None):
        return dict.get(self, key.lower(), def_val)

    def setdefault(self, key, def_val=None):
        return dict.setdefault(self, key.lower(), def_val)

    def update(self, inputdict):
        for k,v in inputdict.items():
            dict.__setitem__(self, k.lower(), v)

    def fromkeys(self, iterable, value=None):
        d = CaseInsensitiveDict()
        for k in iterable:
            dict.__setitem__(d, k.lower(), value)
        return d

    def pop(self, key, def_val=None):
        return dict.pop(self, key.lower(), def_val)
    

length_dict = {'meters':1.0,'m':1.0,
        'centimeters':1e-2,'cm':1e-2,
        'millimeters':1e-3,'mm':1e-3,
        'nanometers':1e-9,'nm':1e-9,
        'micrometers':1e-6,'micron':1e-6,'microns':1e-6,'um':1e-6,
        'kilometers':1e3,'km':1e3,
        'angstroms':1e-10,'A':1e-10,
        }
length_dict = CaseInsensitiveDict(length_dict)

wavelength_dict = length_dict # synonym

frequency_dict = {
        'Hz':1.0,
        'kHz':1e3,
        'MHz':1e6,
        'GHz':1e9,
        'THz':1e12,
        }
frequency_dict = CaseInsensitiveDict(frequency_dict)

velocity_dict = {'meters/second':1.0,'m/s':1.0,
        'kilometers/s':1e3,'km/s':1e3,'kms':1e3,
        'centimeters/s':1e-2,'cm/s':1e-2,'cms':1e-2,
        }
velocity_dict = CaseInsensitiveDict(velocity_dict)

pixel_dict = {'pixel':1,'pixels':1}
pixel_dict = CaseInsensitiveDict(pixel_dict)

conversion_dict = {
        'VELOCITY':velocity_dict,  'Velocity':velocity_dict,  'velocity':velocity_dict,  'velo': velocity_dict, 'VELO': velocity_dict,
        'LENGTH':length_dict,      'Length':length_dict,      'length':length_dict, 
        'WAVELENGTH':length_dict,      'WAVELength':length_dict,      'WAVElength':length_dict, 
        'FREQUENCY':frequency_dict,'Frequency':frequency_dict,'frequency':frequency_dict, 'freq': frequency_dict, 'FREQ': frequency_dict,
        'pixels':pixel_dict,'PIXELS':pixel_dict,
        }
conversion_dict = CaseInsensitiveDict(conversion_dict)

unit_type_dict = {
    'Hz' :'frequency', 'kHz':'frequency', 'MHz':'frequency', 'GHz':'frequency',
    'HZ' :'frequency', 'KHZ':'frequency', 'MHZ':'frequency', 'GHZ':'frequency',
    'hz' :'frequency', 'khz':'frequency', 'mhz':'frequency', 'ghz':'frequency',
    'THz':'frequency', 
    'meters/second':'velocity', 'm/s':'velocity', 'kilometers/s':'velocity',
    'km/s':'velocity', 'kms':'velocity', 'centimeters/s':'velocity',
    'cm/s':'velocity', 'cms':'velocity', 
    'meters':'wavelength','m':'wavelength',
    'centimeters':'wavelength','cm':'wavelength',
    'millimeters':'wavelength','mm':'wavelength',
    'nanometers':'wavelength','nm':'wavelength',
    'micrometers':'wavelength','micron':'wavelength','microns':'wavelength','um':'wavelength',
    'kilometers':'wavelength','km':'wavelength',
    'angstroms':'wavelength','A':'wavelength',
    'unknown':'pixels',
    None: 'pixels',
    }

unit_type_dict = CaseInsensitiveDict(unit_type_dict)

# to-do
# unit_prettyprint = dict([(a,a.replace('angstroms','\\AA')) for a in unit_type_dict.keys() if a is not None])

xtype_dict = {
        'VLSR':'velocity','VRAD':'velocity','VELO':'velocity',
        'VELO-LSR':'velocity',
        'VOPT':'velocity',
        'VHEL':'velocity',
        'VGEO':'velocity',
        'VREST':'velocity',
        'velocity':'velocity',
        'Z':'redshift',
        'FREQ':'frequency',
        'frequency':'frequency',
        'WAV':'wavelength',
        'WAVE':'wavelength',
        'wavelength':'wavelength',
        'lambda':'wavelength',
        # cm^-1 ? 'wavenumber':'wavenumber',
        }

frame_dict = CaseInsensitiveDict({
        'VLSR':'LSRK','VRAD':'LSRK','VELO':'LSRK',
        'VOPT':'LSRK',
        'VELO-LSR':'LSRK',
        'LSRD':'LSRD',
        '-LSR':'LSRK',
        '-HEL':'heliocentric',
        '-BAR':'barycentric',
        'BAR':'barycentric',
        'BARY':'barycentric',
        '-OBS':'obs',
        'VHEL':'heliocentric',
        'VGEO':'geocentric',
        'topo':'geocentric',
        'topocentric':'geocentric',
        'velocity':'LSRK',
        'VREST':'rest',
        'Z':'rest',
        'FREQ':'rest',
        'frequency':'rest',
        'WAV':'rest',
        'WAVE':'rest',
        'wavelength':'rest',
        'lambda':'rest',
        'redshift':'obs'
        })

frame_type_dict = {'LSRK':'velocity','LSRD':'velocity','LSR':'velocity',
        'heliocentric':'velocity','topocentric':'velocity','geocentric':'velocity',
        'rest':'frequency','obs':'frequency','observed':'frequency'}

fits_frame = {'rest':'REST','LSRK':'-LSR','heliocentric':'-HEL','geocentric':'-GEO'}
fits_specsys = {'rest':'REST','LSRK':'LSRK','LSRD':'LSRD','heliocentric':'HEL','geocentric':'GEO'}
fits_type = {'velocity':'VELO','frequency':'FREQ','wavelength':'WAVE','length':'WAVE','redshift':'REDS',
        'Velocity':'VELO','Frequency':'FREQ','Wavelength':'WAVE','Length':'WAVE','Redshift':'REDS'}
convention_suffix = {'radio':'RAD','optical':'OPT','relativistic':'REL','redshift':'RED'}

speedoflight_ms  = np.float64(2.99792458e8) # m/s
speedoflight_kms = np.float64(2.99792458e5) # km/s
c = speedoflight_cms = np.float64(2.99792458e10) # cm/s

astronomical_unit_cm = np.float64(1.49597870e13)

hplanck = h = 6.626068e-27    # Planck's Constant (erg s)
k = kb = kB = 1.3806503e-16   # Boltzmann's Constant (erg / K)
electronmass = me = 9.10938188e-28 # g
electroncharge = 4.80320427e-10 # esu

unitdict = {
        'cgs':{ 'h':6.626068e-27,
            'k':1.3806503e-16,
            'kb':1.3806503e-16,
            'c':2.99792458e10,
            'mh':1.67262158e-24 * 1.00794, # proton mass
            'me':9.10938188e-28,           # electron mass
            'e':4.80320427e-10, # electron charge, esu
            'speed':'cm/s',
            'length':'cm'},
        'mks':{ 'h':6.626068e-34,
            'k':1.3806503e-23,
            'kb':1.3806503e-23,
            'c':2.99792458e8,
            'mh':1.67262158e-27 * 1.00794,
            'me':9.10938188e-31,
            'speed':'m/s',
            'length':'m'}
        }


class SpectroscopicAxis(np.ndarray):
    """
    A Spectroscopic Axis object to store the current units of the spectrum and
    allow conversion to other units and frames.  Typically, units are velocity,
    wavelength, frequency, or redshift.  Wavenumber is also hypothetically
    possible.

    WARNING: If you index a SpectroscopicAxis, the resulting array will be
    a SpectroscopicAxis without a dxarr attribute!  This can result in major problems;
    a workaround is being sought but subclassing numpy arrays is harder than I thought
    """

    def __new__(self, xarr, unit="Hz", frame='rest', frame_offset=0.0,
            frame_offset_units='Hz', xtype=None, refX=None, redshift=None,
            refX_units=None, velocity_convention=None, use128bits=False,
            units=None):
        """
        Make a new spectroscopic axis instance
        Default units Hz
        
        *xarr* [ np.ndarray ]
            An array of X-axis values in whatever unit specified

        *unit* or *units* [ string ]
            Any valid spectroscopic X-axis unit (km/s, Hz, angstroms, etc.)

        *xtype* [ string | None ]
            irrelevant?

        *refX* [ float ]
            Reference frequency/wavelength

        *refX_units* [ string ]
            Units of the reference frequency/wavelength
        
        *frame* [ frame_dict key ]
            **NOT IMPLEMENTED**
            The frame of the axis, e.g. 
                * Topocentric (observatory-center) [up to 0.5 km/s offset from Earth-center]
                * Geocentric (earth-center) 
                * Barycentric (Solar System center) 
                * Heliocentric (Sun-center)
                * Kinematic Local Standard of Rest (LSRK)
                * Dynamic Local Standard of Rest (LSRD)
                * Galactocentric
                * Supergalactic
                * CMB?
                * "Rest" (relative to the rest frequency of the line)
                * Redshifted
            SOME tools for frame shifting are now incorporated into 
            velocity_frames.py
        """
        if use128bits:
            dtype='float128'
        else:
            dtype='float64'
        subarr = np.array(xarr,dtype=dtype)
        subarr = subarr.view(self)

        # allow either plural or singular unit
        if units is not None and units in unit_type_dict and unit=='Hz':
            unit=units

        if unit in unit_type_dict:
            subarr.units = unit
        else:
            raise ValueError('Unit %s not recognized.' % unit)
        if subarr.units is None:
            subarr.units = 'none'
        subarr.frame = frame

        if xtype in xtype_dict:
            subarr.xtype = xtype_dict[xtype]
            subarr.frame = frame_dict[xtype]
        elif subarr.units in unit_type_dict:
            subarr.xtype = unit_type_dict[subarr.units]
        elif type(xtype) is str:
            warnings.warn("Unknown X-axis type in header: %s" % xtype)
            subarr.xtype = xtype
        else:
            if xtype is not None:
                warnings.warn("WARNING: xtype has been specified but was not recognized as a string.")
            subarr.xtype = 'unknown'

        subarr.refX = refX
        if refX_units is None:
            if subarr.units in frequency_dict:
                subarr.refX_units = subarr.units
            else:
                subarr.refX_units = 'Hz'
        else:
            subarr.refX_units = refX_units
        subarr.redshift = redshift
        subarr.wcshead = {}
        if velocity_convention is None:
            if subarr.xtype is not None:
                if 'RAD' in subarr.xtype:
                    subarr.velocity_convention = 'radio'
                elif 'OPT' in subarr.xtype:
                    subarr.velocity_convention = 'optical'
                elif 'REL' in subarr.xtype:
                    subarr.velocity_convention = 'relativistic'
                elif subarr.xtype is 'unknown':
                    subarr.velocity_convention = 'radio' # default
                else:
                    subarr.velocity_convention = 'radio' # default
            else:
                subarr.velocity_convention = 'radio' # default
        else:
            subarr.velocity_convention = velocity_convention

        return subarr

    def __array_finalize__(self,obj):
        """
        Array finalize allows you to take subsets of the array but keep units
        around, e.g.:
        xarr = self[1:20]
        """
        self.units = getattr(obj, 'units', None)
        self.frame = getattr(obj, 'frame', None)
        self.xtype = getattr(obj, 'xtype', None)
        self.refX = getattr(obj, 'refX', None)
        self.refX_units = getattr(obj, 'refX_units', None)
        self.velocity_convention = getattr(obj, 'velocity_convention', None)
        self.redshift = getattr(obj, 'redshift', None)
        self.wcshead = getattr(obj, 'wcshead', None)
        # moved from __init__ - needs to be done whenever viewed
        # (this is slow, though - may be better not to do this)
        if self.shape: # check to make sure non-scalar
            self.dxarr = np.diff(np.array(self))

    def __array_wrap__(self,out_arr,context=None):
        """
        Do this when calling ufuncs
        """
        ret = np.ndarray.__array_wrap__(self, out_arr, context)
        try:
            if len(ret) == 1: # I think this is a bad hack; it used to be unnecessary but a recent update to numpy changed that?
                ret = float(ret) # FORCE scalar
        except TypeError: # sometimes ret has no len
            pass
        return ret

    def __repr__(self):
        if self.shape is ():
            rep = ("SpectroscopicAxis([%r], units=%r, refX=%r, refX_units=%r, frame=%r, redshift=%r, xtype=%r, velocity convention=%r)" %
                (self.__array__(), self.units, self.refX, self.refX_units, self.frame, self.redshift, self.xtype, self.velocity_convention))
        else:
            rep = ("SpectroscopicAxis([%r,...,%r], units=%r, refX=%r, refX_units=%r, frame=%r, redshift=%r, xtype=%r, velocity convention=%r)" %
                (self[0], self[-1], self.units, self.refX, self.refX_units, self.frame, self.redshift, self.xtype, self.velocity_convention))
        return rep

    def __str__(self):
        selfstr =  "SpectroscopicAxis with units %s and range %g:%g." % (
                self.units,self.umin(),self.umax())
        if self.refX is not None:
            selfstr += "Reference is %g %s" % (self.refX, self.refX_units)
        return selfstr

    def _check_consistent_type(self):
        """
        Make sure self.xtype is unit_type_dict[units]
        if this is NOT true, can cause significant errors
        """
        OK = self.xtype == unit_type_dict[self.units]
        if not OK:
            raise InconsistentTypeError("Units: %s Type[units]: %s in %r" % (self.units,unit_type_dict[self.units],self) )

    """ OBSOLETE use convert_to_unit
    def change_xtype(self,new_xtype,**kwargs):
        if new_xtype not in conversion_dict:
            raise ValueError("There is no X-axis type %s" % new_xtype)
        if conversion_dict[self.xtype] is velocity_dict:
            if conversion_dict[new_xtype] is frequency_dict:
                self.velocity_to_frequency(**kwargs)
            elif conversion_dict[new_xtype] is wavelength_dict:
                self.velocity_to_frequency()
                self.frequency_to_wavelength(**kwargs)
        elif conversion_dict[self.xtype] is frequency_dict:
            if conversion_dict[new_xtype] is velocity_dict:
                self.frequency_to_velocity(**kwargs)
            elif conversion_dict[new_xtype] is wavelength_dict:
                self.frequency_to_wavelength(**kwargs)
        elif conversion_dict[self.xtype] is wavelength_dict:
            if conversion_dict[new_xtype] is velocity_dict:
                self.wavelength_to_frequency()
                self.frequency_to_velocity(**kwargs)
            elif conversion_dict[new_xtype] is wavelength_dict:
                self.wavelength_to_frequency(**kwargs)
        else:
            raise ValueError("Conversion to xtype %s was not recognized." % xtype)
    """

    def umax(self, units=None):
        """
        Return the maximum value of the SpectroscopicAxis.  If units specified,
        convert to those units first
        """

        if units is not None:
            # could be reversed
            return np.max([self.coord_to_x(self.max(), units),self.coord_to_x(self.min(),units)])
        else: 
            return self.max()

    def umin(self, units=None):
        """
        Return the minimum value of the SpectroscopicAxis.  If units specified,
        convert to those units first
        """

        if units is not None:
            # could be reversed
            return np.min([self.coord_to_x(self.max(), units),self.coord_to_x(self.min(),units)])
        else: 
            return self.min()

    def x_to_pix(self, xval):
        """
        Given an X coordinate in SpectroscopicAxis' units, return the corresponding pixel number
        """
        nearest_pix = np.argmin(np.abs(self-xval))
        return nearest_pix

    def in_range(self, xval):
        """
        Given an X coordinate in SpectroscopicAxis' units, return whether the pixel is in range
        """
        return bool((xval > self.min()) * (xval < self.max()))

    def x_to_coord(self, xval, xunit, verbose=False):
        """
        Given a wavelength/frequency/velocity, return the value in the SpectroscopicAxis's units
        e.g.:
        xarr.units = 'km/s'
        xarr.refX = 5.0 
        xarr.refX_units = GHz
        xarr.x_to_coord(5.1,'GHz') == 6000 # km/s
        """
        #if xunit in wavelength_dict: # shortcut - convert wavelength to freq
        #    xval = wavelength_to_frequency(xval, xunit, 'Hz')
        #    xunit = 'Hz'

        if unit_type_dict[self.units] == unit_type_dict[xunit]:
            if verbose: print "Input units of same type as output units."
            return xval * conversion_dict[unit_type_dict[xunit]][xunit] / conversion_dict[unit_type_dict[self.units]][self.units]
        elif xunit in velocity_dict:
            if verbose: print "Requested units are Velocity"
            if self.units in frequency_dict:
                return velocity_to_frequency(xval, xunit,
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        frequency_units=self.units,
                        convention=self.velocity_convention)
            elif self.units in wavelength_dict:
                FREQ = velocity_to_frequency(xval, xunit,
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        frequency_units='Hz',
                        convention=self.velocity_convention)
                return frequency_to_wavelength(FREQ, 'Hz',
                        wavelength_units=self.units)
        elif xunit in frequency_dict:
            if verbose: print "Requested units are Frequency"
            if self.units in velocity_dict:
                return frequency_to_velocity(xval, xunit,
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        velocity_units=self.units,
                        convention=self.velocity_convention)
            elif self.units in wavelength_dict:
                return frequency_to_wavelength(xval, xunit,
                        wavelength_units=self.units)
        elif xunit in wavelength_dict:
            if verbose: print "Requested units are Frequency"
            if self.units in velocity_dict:
                return wavelength_to_velocity(xval, xunit,
                        center_wavelength=self.refX,
                        center_wavelength_units=self.refX_units,
                        velocity_units=self.units,
                        convention=self.velocity_convention)
            elif self.units in frequency_dict:
                return wavelength_to_frequency(xval, xunit,
                        frequency_units=self.units)
        else:
            raise ValueError("Units not recognized.")

    def coord_to_x(self, xval, xunit):
        """
        Given an X-value assumed to be in the coordinate axes, return that
        value converted to xunit
        e.g.:
        xarr.units = 'km/s'
        xarr.refX = 5.0 
        xarr.refX_units = GHz
        xarr.coord_to_x(6000,'GHz') == 5.1 # GHz
        """
        if unit_type_dict[self.units] == unit_type_dict[xunit]:
            return xval / conversion_dict[unit_type_dict[xunit]][xunit] * conversion_dict[unit_type_dict[self.units]][self.units]


        if xunit in velocity_dict:
            if self.units in frequency_dict:
                return frequency_to_velocity(xval, self.units,
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        velocity_units=xunit,
                        convention=self.velocity_convention)
            elif self.units in wavelength_dict:
                FREQ = wavelength_to_frequency(xval, self.units, frequency_units='Hz')
                return frequency_to_velocity(FREQ, 'Hz',
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        velocity_units=xunit,
                        convention=self.velocity_convention)
        elif xunit in frequency_dict:
            if self.units in velocity_dict:
                return velocity_to_frequency(xval, self.units,
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        frequency_units=xunit,
                        convention=self.velocity_convention)
            elif self.units in wavelength_dict:
                return wavelength_to_frequency(xval, self.units, frequency_units=xunit)
        elif xunit in wavelength_dict:
            if self.units in velocity_dict:
                FREQ = velocity_to_frequency(xval, self.units,
                        center_frequency=self.refX,
                        center_frequency_units=self.refX_units,
                        frequency_units='Hz',
                        convention=self.velocity_convention)
                return frequency_to_wavelength(FREQ, 'Hz', wavelength_units=xunit)
            elif self.units in frequency_dict:
                return frequency_to_wavelength(xval, self.units, wavelength_units=xunit)
        else:
            raise ValueError("Units not recognized.")

    def convert_to_unit(self, unit, **kwargs):
        """
        Return the X-array in the specified units without changing it
        Uses as_unit for the conversion, but changes internal values rather
        than returning them.
        """
        self[:] = self.as_unit(unit, **kwargs)
        
        if unit in velocity_dict:
            self.xtype = "Velocity"
        elif unit in frequency_dict:
            self.xtype = "Frequency"
        elif unit in wavelength_dict:
            self.xtype = "Wavelength"

        if unit in unit_type_dict and unit not in (None, 'unknown'):
            self.units = unit
        self.dxarr = np.diff(self)

    def as_unit(self, unit, frame=None, quiet=True, center_frequency=None,
            center_frequency_units=None, debug=False, **kwargs):
        """
        Convert the spectrum to the specified units.  This is a wrapper function
        to convert between frequency/velocity/wavelength and simply change the 
        units of the X axis.  Frame conversion is... not necessarily implemented.

        *unit* [ string ] 
            What unit do you want to 'view' the array as?
            None returns the x-axis unchanged (NOT a copy!)

        *frame* [ None ]
            NOT IMPLEMENTED.  When it is, it will allow you to convert between
            LSR, topocentric, heliocentric, rest, redshifted, and whatever other
            frames we can come up with.  Right now the main holdup is finding a 
            nice python interface to an LSR velocity calculator... and motivation.
 
        *center_frequency* [ None | float ]
        *center_frequency_units* [ None | string ]
            If converting between velocity and any other spectroscopic type,
            need to specify the central frequency around which that velocity is
            calculated.
            I think this can also accept wavelengths....
        """

        if unit in (None,'none'):
            return self
        elif unit in pixel_dict:
            return np.arange(self.shape[0])
        elif unit not in conversion_dict[self.xtype]:
            change_xtype = True
            change_units = True
        elif unit != self.units: 
            change_xtype = False
            change_units = True
        else: 
            change_xtype = False
            change_units = False
        if frame is not None and frame != self.frame: change_frame = True
        else: change_frame = False

        if center_frequency is None:
            center_frequency = self.refX
        if center_frequency_units is None:
            center_frequency_units = self.refX_units

        if change_xtype:
            if unit in velocity_dict:
                if debug: print "Converting to velocity"
                if conversion_dict[self.xtype] is frequency_dict:
                    if debug: print "Converting frequency to velocity"
                    newxarr = frequency_to_velocity(self,self.units,
                            center_frequency=center_frequency,
                            center_frequency_units=center_frequency_units,
                            velocity_units=unit, convention=self.velocity_convention)
                    newxtype = "Velocity"
                    newunit = unit
                elif conversion_dict[self.xtype] is wavelength_dict:
                    if debug: print "Converting wavelength to velocity"
                    freqx = wavelength_to_frequency(self, self.units) 
                    cf = wavelength_to_frequency(center_frequency, center_frequency_units)
                    cfu = 'GHz'
                    newxarr = frequency_to_velocity(freqx, 'GHz',
                            center_frequency=cf,
                            center_frequency_units=cfu,
                            velocity_units=unit, convention=self.velocity_convention)
                    newxtype = "Velocity"
                    newunit = unit
                else: 
                    raise ValueError("Could not convert from %s to %s" % (self.units,unit))
            elif unit in frequency_dict:
                if debug: print "Converting to frequency"
                if conversion_dict[self.xtype] is velocity_dict:
                    if debug: print "Converting velocity to frequency"
                    newxarr = velocity_to_frequency(self,self.units,
                            center_frequency=center_frequency,
                            center_frequency_units=center_frequency_units,
                            frequency_units=unit, convention=self.velocity_convention)
                    newxtype = "Frequency"
                    newunit = unit
                elif conversion_dict[self.xtype] is wavelength_dict:
                    if debug: print "Converting wavelength to frequency"
                    newxarr = wavelength_to_frequency(self, self.units,
                            frequency_units=unit)
                    newxtype = "Frequency"
                    newunit = unit
                else: 
                    raise ValueError("Could not convert from %s to %s" % (self.units,unit))
            elif unit in wavelength_dict:
                if debug: print "Converting to wavelength"
                if conversion_dict[self.xtype] is velocity_dict:
                    if debug: print "Converting velocity to wavelength"
                    if center_frequency_units in frequency_dict:
                        freqx = velocity_to_frequency(self, self.units,
                                center_frequency=center_frequency,
                                center_frequency_units=center_frequency_units,
                                frequency_units='Hz',
                                convention=self.velocity_convention)
                        newxarr = frequency_to_wavelength(freqx, 'Hz',
                                wavelength_units=unit)
                    elif center_frequency_units in wavelength_dict:
                        newxarr = velocity_to_wavelength(self, self.units,
                                center_wavelength=center_frequency,
                                center_wavelength_units=center_frequency_units,
                                wavelength_units=unit,
                                convention=self.velocity_convention)
                    newxtype = "Wavelength"
                    newunit = unit
                elif conversion_dict[self.xtype] is frequency_dict:
                    if debug: print "Converting frequency to wavelength"
                    newxarr = frequency_to_wavelength(self, self.units,
                            wavelength_units=unit)
                    newxtype = "Wavelength"
                    newunit = unit
                else: 
                    raise ValueError("Could not convert from %s to %s" % (self.units,unit))
            else:
                warnings.warn("Could not convert from %s to %s" % (self.units,unit))
        else:
            newxtype = self.xtype
            newxarr = self
            newunit = self.units

        if debug: 
            print "Converted from %s:%s to %s:%s" % (self.xtype, self.units, newxtype, newunit)

        # re-check whether units need to be changed; it is possible that change_xtype left you 
        # with the correct units
        if unit != newunit and change_units:
            conversion_factor = conversion_dict[newxtype][newunit] / conversion_dict[newxtype][unit] 
            if not quiet: print "Converting units from %s to %s" % (newunit,unit)
            newxarr = newxarr*conversion_factor
            newunit = unit

        if change_frame and not quiet:
            if not quiet: print "Conversion from frame %s to %s is not yet supported" % (self.frame,frame)

        if not change_units and not change_xtype and not change_frame:
            if not quiet: print "Already in desired units, X-type, and frame"

        if not quiet:
            print "Converted to %s (%s)" % (newxtype,newunit)

        # this should be implemented but requires a callback to spectrum...
        #if replot:
        #    self.spectrum.plotter(reset_xlimits=True)

        newxarr.units = newunit
        newxarr.xtype = newxtype

        return newxarr

    def change_frame(self, frame):
        """
        Change velocity frame
        """
        if self.frame == frame:
            return
        
        if frame in frame_type_dict:
            self[:] = self.in_frame(frame)
            self.dxarr = np.diff(self)
            self.frame = frame

    def make_dxarr(self, coordinate_location='center'):
        """
        Create a "delta-x" array corresponding to the X array.

        *coordinate_location* [ 'left', 'center', 'right' ]
            Does the coordinate mark the left, center, or right edge of the
            pixel?  If 'center' or 'left', the *last* pixel will have the same
            dx as the second to last pixel.  If right, the *first* pixel will
            have the same dx as the second pixel.

        """

    def in_frame(self, frame):
        """
        Return a shifted xaxis
        """
        if self.frame in ('obs','observed') and frame in ('rest',):
            if self.redshift is not None:
                return self/(1.0+self.redshift)
            else:
                warnings.warn("WARNING: Redshift is not specified, so no shift will be done.")
        elif self.frame in ('rest',) and frame in ('obs','observed'):
            if self.redshift is not None:
                return self*(1.0+self.redshift)
            else:
                warnings.warn("WARNING: Redshift is not specified, so no shift will be done.")
        else:
            print "Frame shift from %s to %s is not defined or implemented." % (self.frame, frame)
            return self

    def x_in_frame(self, xx, frame):
        """
        Return the value 'x' shifted to the target frame
        """
        return self.in_frame(frame)[self.x_to_pix(xx)]

    def cdelt(self, tolerance=1e-8):
        """
        Return the channel spacing if channels are linear
        """
        if not hasattr(self,'dxarr'): # if cropping happens...
            self.dxarr = np.diff(self)
        if abs(self.dxarr.max()-self.dxarr.min())/abs(self.dxarr.min()) < tolerance:
            return self.dxarr.mean().flat[0]

    def _make_header(self, tolerance=1e-8):
        """
        Generate a set of WCS parameters for the X-array
        """
        self.dxarr = np.diff(self)

        if self.wcshead is None:
            self.wcshead = {}

        self.wcshead['CUNIT1'] = self.units
        if fits_type[self.xtype] == 'VELO' and self.velocity_convention is not None:
            ctype = 'V'+convention_suffix[self.velocity_convention]
        else:
            ctype = fits_type[self.xtype]
        self.wcshead['CTYPE1'] = ctype+fits_frame[self.frame]
        self.wcshead['SPECSYS'] = fits_specsys[self.frame]
        self.wcshead['RESTFRQ'] = self.refX

        # check to make sure the X-axis is linear
        cdelt = self.cdelt(tolerance=tolerance)
        if cdelt is None:
            self.wcshead['CDELT1'] = None
            self.wcshead['CRPIX1'] = None
            self.wcshead['CRVAL1'] = None
            return False
        else:
            self.wcshead['CDELT1'] = cdelt
            self.wcshead['CRVAL1'] = self[0]
            self.wcshead['CRPIX1'] = 1.0
            return True

class SpectroscopicAxes(SpectroscopicAxis):
    """
    Counterpart to Spectra: takes a list of SpectroscopicAxis's and
    concatenates them while checking for consistency and maintaining
    header parameters
    """

    def __new__(self, axislist, frame='rest', xtype=None, refX=None,
            redshift=None):

        if type(axislist) is not list:
            raise TypeError("SpectroscopicAxes must be initiated with a list of SpectroscopicAxis objects")

        units = axislist[0].units
        xtype = axislist[0].xtype
        frame = axislist[0].frame
        redshift = axislist[0].redshift
        velocity_convention = axislist[0].velocity_convention
        for ax in axislist:
            if ax.xtype != xtype:
                try: 
                    ax.change_xtype(xtype)
                except:
                    ValueError("Axis had wrong xtype and could not be converted.")
            if ax.units != units or ax.frame != frame:
                try:
                    ax.convert_to_unit(units,frame=frame)
                except:
                    ValueError("Axis had wrong units and could not be converted.")

        subarr = np.concatenate([ax for ax in axislist])
        subarr = subarr.view(self)
        subarr.units = units
        subarr.xtype = xtype
        subarr.frame = frame
        subarr.redshift = redshift
        subarr.velocity_convention = velocity_convention

        # if all the spectra have the same reference frequency, there is one common refX
        # else, refX is undefined and velocity transformations should not be done
        refXs_diff = np.sum([axislist[0].refX != ax.refX for ax in axislist])
        if refXs_diff > 0:
            subarr.refX = None
            subarr.refX_units = None
        else:
            subarr.refX = axislist[0].refX
            subarr.refX_units = axislist[0].refX_units

        return subarr

class EchelleAxes(SpectroscopicAxis):
    """
    Counterpart to Spectra: takes a list of SpectroscopicAxis's and
    stacks them while checking for consistency and maintaining
    header parameters
    """

    def __new__(self, axislist, frame='rest', xtype=None, refX=None,
            redshift=None):

        if type(axislist) is not list:
            raise TypeError("SpectroscopicAxes must be initiated with a list of SpectroscopicAxis objects")

        units = axislist[0].units
        xtype = axislist[0].xtype
        frame = axislist[0].frame
        redshift = axislist[0].redshift
        velocity_convention = axislist[0].velocity_convention
        for ax in axislist:
            if ax.xtype != xtype:
                try: 
                    ax.change_xtype(xtype)
                except:
                    ValueError("Axis had wrong xtype and could not be converted.")
            if ax.units != units or ax.frame != frame:
                try:
                    ax.convert_to_unit(units,frame=frame)
                except:
                    ValueError("Axis had wrong units and could not be converted.")

        subarr = np.array(axislist)
        subarr = subarr.view(self)
        subarr.units = units
        subarr.xtype = xtype
        subarr.frame = frame
        subarr.redshift = redshift
        subarr.velocity_convention = velocity_convention

        # if all the spectra have the same reference frequency, there is one common refX
        # else, refX is undefined and velocity transformations should not be done
        refXs_diff = np.sum([axislist[0].refX != ax.refX for ax in axislist])
        if refXs_diff > 0:
            subarr.refX = None
            subarr.refX_units = None
        else:
            subarr.refX = axislist[0].refX
            subarr.refX_units = axislist[0].refX_units

        return subarr

def velocity_to_frequency(velocities, input_units, center_frequency=None,
        center_frequency_units=None, frequency_units='Hz',
        convention='radio'):
    """
    Conventions defined here:
    http://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html
    * Radio 	V = c (f0 - f)/f0 	f(V) = f0 ( 1 - V/c )
    * Optical 	V = c (f0 - f)/f 	f(V) = f0 ( 1 + V/c )^-1
    * Redshift 	z = (f0 - f)/f 	f(V) = f0 ( 1 + z )-1
    * Relativistic 	V = c (f02 - f 2)/(f02 + f 2) 	f(V) = f0 { 1 - (V/c)2}1/2/(1+V/c) 
    """
    if input_units in frequency_dict:
        print "Already in frequency units (%s)" % input_units
        return velocities
    if center_frequency is None:
        raise ValueError("Cannot convert velocity to frequency without specifying a central frequency.")
    if frequency_units not in frequency_dict:
        raise ValueError("Bad frequency units: %s" % (frequency_units))

    velocity_ms = velocities / velocity_dict['m/s'] * velocity_dict[input_units]
    if convention == 'radio':
        freq = (velocity_ms / speedoflight_ms - 1.0) * center_frequency * -1
    elif convention == 'optical':
        freq = (velocity_ms / speedoflight_ms + 1.0)**-1 * center_frequency
    elif convention == 'relativistic':
        freq = (-(velocity_ms / speedoflight_ms)**2 + 1.0)**0.5 / (1.0 + velocity_ms/speedoflight_ms) * center_frequency
    else:
        raise ValueError('Convention "%s" is not allowed.' % (convention))
    frequencies = freq / frequency_dict[frequency_units] * frequency_dict[center_frequency_units]
    return frequencies

def frequency_to_velocity(frequencies, input_units, center_frequency=None,
        center_frequency_units=None, velocity_units='m/s',
        convention='radio'):
    """
    Conventions defined here:
    http://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html
    * Radio 	V = c (f0 - f)/f0 	f(V) = f0 ( 1 - V/c )
    * Optical 	V = c (f0 - f)/f 	f(V) = f0 ( 1 + V/c )-1
    * Redshift 	z = (f0 - f)/f 	f(V) = f0 ( 1 + z )-1
    * Relativistic 	V = c (f02 - f 2)/(f02 + f 2) 	f(V) = f0 { 1 - (V/c)2}1/2/(1+V/c) 
    """
    if input_units in velocity_dict:
        print "Already in velocity units (%s)" % input_units
        return frequencies
    if center_frequency is None:
        raise ValueError("Cannot convert frequency to velocity without specifying a central frequency.")
    if center_frequency_units not in frequency_dict:
        raise ValueError("Bad frequency units: %s" % (center_frequency_units))
    if velocity_units not in velocity_dict:
        raise ValueError("Bad velocity units: %s" % (velocity_units))

    frequency_hz = frequencies / frequency_dict['Hz'] * frequency_dict[input_units]
    center_frequency_hz = center_frequency / frequency_dict['Hz'] * frequency_dict[center_frequency_units]

    # the order is very ugly because otherwise, if scalar, the spectroscopic axis attributes won't be inherited
    if convention == 'radio':
        velocity = ( frequency_hz - center_frequency_hz ) / center_frequency_hz * speedoflight_ms * -1
    elif convention == 'optical':
        velocity = ( frequency_hz - center_frequency_hz ) / frequency_hz * speedoflight_ms * -1
    elif convention == 'relativistic':
        velocity = ( frequency_hz**2 - center_frequency_hz**2 ) / ( center_frequency_hz**2 + frequency_hz )**2 * speedoflight_ms * -1
    else:
        raise ValueError('Convention "%s" is not allowed.' % (convention))
    velocities = velocity * velocity_dict['m/s'] / velocity_dict[velocity_units]

    return velocities

def frequency_to_wavelength(frequencies, input_units, wavelength_units='um'):
    """
    Simple conversion from frequency to wavelength:
    lambda = c / nu
    """
    if input_units in wavelength_dict:
        print "Already in wavelength units (%s)" % input_units
        return
    if wavelength_units not in length_dict:
        raise ValueError("Wavelength units %s not valid" % wavelength_units)

    if input_units not in frequency_dict:
        raise AttributeError("Cannot convert from frequency unless units are already frequency.")

    wavelengths = speedoflight_ms / ( frequencies * frequency_dict[input_units] ) / length_dict[wavelength_units]

    return wavelengths

def wavelength_to_frequency(wavelengths, input_units, frequency_units='GHz'):
    """
    Simple conversion from frequency to wavelength:
    nu = c / lambda
    """
    if input_units in frequency_dict:
        print "Already in frequency units (%s)" % input_units
        return wavelengths
    if frequency_units not in frequency_dict:
        raise ValueError("Frequency units %s not valid" % wavelength_units)

    if input_units not in length_dict:
        raise AttributeError("Cannot convert from wavelength unless units are already wavelength.")

    frequencies = speedoflight_ms / ( wavelengths * length_dict[input_units] ) / frequency_dict[frequency_units]

    return frequencies

def velocity_to_wavelength(velocities, input_units, center_wavelength=None,
        center_wavelength_units=None, wavelength_units='meters',
        convention='optical'):
    """
    Conventions defined here:
    http://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html
    * Radio 	V = c (c/l0 - c/l)/(c/l0) 	f(V) = (c/l0) ( 1 - V/c )
    * Optical 	V = c ((c/l0) - (c/l))/(c/l) 	f(V) = (c/l0) ( 1 + V/c )^-1
    * Redshift 	z = ((c/l0) - (c/l))/(c/l) 	f(V) = (c/l0) ( 1 + z )-1
    * Relativistic 	V = c ((c/l0)^2 - (c/l)^2)/((c/l0)^2 + (c/l)^2) 	f(V) = (c/l0) { 1 - (V/c)2}1/2/(1+V/c) 
    """
    if input_units in wavelength_dict:
        print "Already in wavelength units (%s)" % input_units
        return velocities
    if center_wavelength is None:
        raise ValueError("Cannot convert velocity to wavelength without specifying a central wavelength.")
    if wavelength_units not in wavelength_dict:
        raise ValueError("Bad wavelength units: %s" % (wavelength_units))

    velocity_ms = velocities / velocity_dict['m/s'] * velocity_dict[input_units]
    center_frequency = speedoflight_ms / center_wavelength
    if convention == 'radio':
        wav = (velocity_ms / speedoflight_ms - 1.0) * center_frequency * -1
    elif convention == 'optical':
        wav = (velocity_ms / speedoflight_ms + 1.0)**-1 * center_frequency
    elif convention == 'relativistic':
        wav = (-(velocity_ms / speedoflight_ms)**2 + 1.0)**0.5 / (1.0 + velocity_ms/speedoflight_ms) * center_frequency
    else:
        raise ValueError('Convention "%s" is not allowed.' % (convention))
    wavelengths = wav / wavelength_dict[wavelength_units] * wavelength_dict[center_wavelength_units]
    return wavelengths

def wavelength_to_velocity(wavelengths, input_units, center_wavelength=None,
        center_wavelength_units=None, velocity_units='m/s',
        convention='optical'):
    """
    Conventions defined here:
    http://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html
    * Radio 	V = c (c/l0 - c/l)/(c/l0) 	f(V) = (c/l0) ( 1 - V/c )
    * Optical 	V = c ((c/l0) - f)/f 	f(V) = (c/l0) ( 1 + V/c )^-1
    * Redshift 	z = ((c/l0) - f)/f 	f(V) = (c/l0) ( 1 + z )-1
    * Relativistic 	V = c ((c/l0)^2 - f^2)/((c/l0)^2 + f^2) 	f(V) = (c/l0) { 1 - (V/c)2}1/2/(1+V/c) 
    """
    if input_units in velocity_dict:
        print "Already in velocity units (%s)" % input_units
        return wavelengths
    if center_wavelength is None:
        raise ValueError("Cannot convert wavelength to velocity without specifying a central wavelength.")
    if center_wavelength_units not in wavelength_dict:
        raise ValueError("Bad wavelength units: %s" % (center_wavelength_units))
    if velocity_units not in velocity_dict:
        raise ValueError("Bad velocity units: %s" % (velocity_units))

    wavelength_m = wavelengths / wavelength_dict['meters'] * wavelength_dict[input_units]
    center_wavelength_m = center_wavelength / wavelength_dict['meters'] * wavelength_dict[center_wavelength_units]
    frequency_hz = speedoflight_ms / wavelength_m
    center_frequency_hz = speedoflight_ms / center_wavelength_m

    # the order is very ugly because otherwise, if scalar, the spectroscopic axis attributes won't be inherited
    if convention == 'radio':
        velocity = ( frequency_hz - center_frequency_hz ) / center_frequency_hz * speedoflight_ms * -1
    elif convention == 'optical':
        velocity = ( frequency_hz - center_frequency_hz ) / frequency_hz * speedoflight_ms * -1
    elif convention == 'relativistic':
        velocity = ( frequency_hz**2 - center_frequency_hz**2 ) / ( center_frequency_hz**2 + frequency_hz )**2 * speedoflight_ms * -1
    else:
        raise ValueError('Convention "%s" is not allowed.' % (convention))
    velocities = velocity * velocity_dict['m/s'] / velocity_dict[velocity_units]

    return velocities

class InconsistentTypeError(Exception):
    pass

def parse_veldef(veldef):
    """
    Try to parse an 8-character FITS veldef
    """

    vconv_dict = {'OPTI':'optical','RADI':'radio','RELA':'relativistic'}

    vconv = veldef[:4]
    velocity_convention = vconv_dict[vconv]

    frame = veldef[4:]
    frame_type = frame_dict[frame]

    return velocity_convention,frame_type




#!/usr/bin/env python

import os
import shutil
import sys
from distutils.core import setup, Command

with open('README.txt') as file:
    long_description = file.read()

with open('CHANGES') as file:
    long_description += file.read()

try:  # Python 3.x
    from distutils.command.build_py import build_py_2to3 as build_py
except ImportError:  # Python 2.x
    from distutils.command.build_py import build_py


from pyspeckit import __version__ as version_base
tagname = "pyspeckit_%s" % (version_base)

if os.path.exists(".hg"):
    # if the installed version is from a mercurial clone, get the "tip", else get the latest release
    try:
        import subprocess
        currentversion = subprocess.Popen(["hg","id","--num"],stdout=subprocess.PIPE).communicate()[0].strip().strip("+")
        tags = subprocess.Popen(["hg","tags"],stdout=subprocess.PIPE).communicate()[0].split()
        tagdict = dict((tags[i],tags[i+1].split(':')[0]) for i in range(0, len(tags), 2))
        tagdict.pop('tip')

        # set "default" to be the tip
        version = version_base+"hg"+currentversion
        download_url = "https://bitbucket.org/pyspeckit/pyspeckit.bitbucket.org/get/tip.tar.gz"
        
        # find out if current version matches any tags.  If so, set version to be that
        for k,v in tagdict.iteritems():
            if int(currentversion) in [int(v),int(v)+1]:
                version = k.strip("pyspeckit_")
                download_url = "https://bitbucket.org/pyspeckit/pyspeckit.bitbucket.org/get/%s.tar.gz" % (k)
    except:
        # is this bad practice?  I don't care if it's an import error, attribute error, or value error...
        version = version_base
        download_url = "https://bitbucket.org/pyspeckit/pyspeckit.bitbucket.org/get/%s.tar.gz" % (tagname)
else:
    version = version_base
    download_url = "https://bitbucket.org/pyspeckit/pyspeckit.bitbucket.org/get/%s.tar.gz" % (tagname)


class PyTest(Command):

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        errno = subprocess.call([sys.executable, 'tests/runtests.py', '-q'])
        raise SystemExit(errno)

setup(name='pyspeckit',
      version=version,
      description='Toolkit for fitting and manipulating spectroscopic data in python',
      long_description=long_description,
      author=['Adam Ginsburg','Jordan Mirocha'],
      author_email=['adam.g.ginsburg@gmail.com', 'mirochaj@gmail.com',
          'pyspeckit@gmail.com'], 
      url='https://pyspeckit.bitbucket.org/',
      download_url=download_url,
      packages=['pyspeckit', 'pyspeckit.spectrum', 'pyspeckit.spectrum.models',
          'pyspeckit.spectrum.readers', 'pyspeckit.spectrum.writers',
          'pyspeckit.spectrum.speclines', 'pyspeckit.cubes',
          'pyspeckit.wrappers','pyspeckit.mpfit','pyspeckit.parallel_map'],
      package_dir={'pyspeckit.spectrum.speclines':'pyspeckit/spectrum/speclines',
          'mpfit':'mpfit'}, 
      package_data={'pyspeckit.spectrum.speclines':[],
          '':['pyspeckit/config_default']},
      requires=['matplotlib (>=1.1.0)','numpy (>=1.4.1)','ordereddict'],
      cmdclass={'build_py': build_py, 'test': PyTest},
      classifiers=[
                   "Development Status :: 3 - Alpha",
                   "Programming Language :: Python",
                   "License :: OSI Approved :: MIT License",
                  ],
      
     )

# These lines break the configuration:
# # Copy default config file to directory $HOME/.pyspeckit
# HOME = os.environ.get('HOME')
# CWD = os.getcwd()
# if not os.path.exists('%s/.pyspeckit' % HOME):
#     os.mkdir('%s/.pyspeckit' % HOME)
#     
# if not os.path.exists('%s/.pyspeckit/config' % HOME):
#     shutil.copyfile('%s/pyspeckit/config_default' % CWD, '%s/.pyspeckit/config' % HOME)
